# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import unicode_literals
import re
import zipfile
import pytest
from grizzly.reduce.args import ReducerArgs, ReducerFuzzManagerIDArgs, ReducerFuzzManagerIDQualityArgs
from grizzly.reduce import reduce, crash, bucket
from grizzly.common import reporter
from .test_common import BaseFakeReporter, TestMainReductionJob, TestMainCrashReductionJob


def test_parse_args(capsys, tmp_path):
    "test that grizzly.reduce args are accepted and validated"
    exe = tmp_path / "binary"
    inp = tmp_path / "input"

    # missing arg tests
    with pytest.raises(SystemExit):
        ReducerArgs().parse_args([])
    _, err = capsys.readouterr()
    with pytest.raises(SystemExit):
        ReducerArgs().parse_args([str(exe)])
    _, err = capsys.readouterr()

    # invalid binary tests
    with pytest.raises(SystemExit):
        ReducerArgs().parse_args([str(exe), str(inp)])
    _, err = capsys.readouterr()
    assert "error: file not found: %r" % (str(exe),) in err
    exe.mkdir()
    with pytest.raises(SystemExit):
        ReducerArgs().parse_args([str(exe), str(inp)])
    _, err = capsys.readouterr()
    assert "error: file not found: %r" % (str(exe),) in err
    exe.rmdir()
    exe.touch()

    # invalid input tests
    with pytest.raises(SystemExit):
        ReducerArgs().parse_args([str(exe), str(inp)])
    _, err = capsys.readouterr()
    assert "error: %r does not exist" % (str(inp),) in err
    inp.touch()
    with pytest.raises(SystemExit):
        ReducerArgs().parse_args([str(exe), str(inp)])
    _, err = capsys.readouterr()
    assert "error: Testcase should be a folder, zip, or html file" in err
    inp.unlink()
    with pytest.raises(SystemExit):
        ReducerFuzzManagerIDArgs().parse_args([str(exe), str(inp)])
    _, err = capsys.readouterr()
    assert "invalid int value" in err

    # valid binary & inputs
    (tmp_path / "input.zip").touch()
    zipf = tmp_path / "input.zip"
    ReducerArgs().parse_args([str(exe), str(zipf)])
    zipf.unlink()
    inp.mkdir()
    (inp / "test_info.txt").touch()
    ReducerArgs().parse_args([str(exe), str(inp)])
    ReducerFuzzManagerIDArgs().parse_args([str(exe), '123'])

    # sig/environ tests
    fname = tmp_path / "file.txt"
    for arg in ("--sig", "--environ"):
        with pytest.raises(SystemExit):
            ReducerArgs().parse_args([str(exe), str(inp), arg, str(fname)])
        _, err = capsys.readouterr()
        assert "error: file not found: %r" % (str(fname),) in err
        fname.mkdir()
        with pytest.raises(SystemExit):
            ReducerArgs().parse_args([str(exe), str(inp), arg, str(fname)])
        _, err = capsys.readouterr()
        assert "error: file not found: %r" % (str(fname),) in err
        fname.rmdir()
        fname.touch()
        ReducerArgs().parse_args([str(exe), str(inp), arg, str(fname)])
        fname.unlink()

    # repeat/min-crashes tests
    for arg in ("--repeat", "--min-crashes"):
        with pytest.raises(SystemExit):
            ReducerArgs().parse_args([str(exe), str(inp), arg, "abc"])
        with pytest.raises(SystemExit):
            ReducerArgs().parse_args([str(exe), str(inp), arg, "-1"])
        _, err = capsys.readouterr()
        assert "'%s' value must be positive" % (arg,) in err
        with pytest.raises(SystemExit):
            ReducerArgs().parse_args([str(exe), str(inp), arg, "0"])
        _, err = capsys.readouterr()
        assert "'%s' value must be positive" % (arg,) in err
        ReducerArgs().parse_args([str(exe), str(inp), arg, "1"])
        ReducerArgs().parse_args([str(exe), str(inp), arg, "10"])


def test_main(mocker, tmp_path):  # noqa pylint: disable=redefined-outer-name
    "simple test that main functions"

    (tmp_path / "binary").touch()
    exe = tmp_path / "binary"
    (tmp_path / "input").mkdir()
    inp = tmp_path / "input"
    (inp / "test_info.txt").write_text("landing page: test.html")
    (inp / "test.html").write_text("fluff\nrequired\n")
    args = ReducerArgs().parse_args([str(exe), str(inp)])
    # uses the job fixture from test_reduce which reduces testcases to the string "required\n"
    mocker.patch("grizzly.reduce.reduce.load_target")
    assert TestMainReductionJob.main(args) == 0


def test_main_prefs(mocker, tmp_path):
    "cmd line prefs should override prefs in the testcase"
    run_called = [0]

    class MyReductionJob(TestMainReductionJob):
        __slots__ = []

        def __init__(self, *_args, **_kwds):
            super(MyReductionJob, self).__init__(tmp_path, create_binary=False)

        def run(self, *args, **kwds):
            result = super(MyReductionJob, self).run(*args, **kwds)
            with open(self.target.prefs) as prefs_fp:
                assert "main prefs" == prefs_fp.read()
            run_called[0] += 1
            return result

    (tmp_path / "binary").touch()
    exe = tmp_path / "binary"
    (tmp_path / "input").mkdir()
    inp = tmp_path / "input"
    (inp / "test_info.txt").write_text("landing page: test.html")
    (inp / "prefs.js").write_text("test prefs")
    (inp / "test.html").write_text("fluff\nrequired\n")
    (tmp_path / "prefs.js").write_text("main prefs")
    args = ReducerArgs().parse_args([str(exe), str(inp),
                                     "-p", str(tmp_path / "prefs.js")])
    mocker.patch("grizzly.reduce.reduce.load_target")
    assert MyReductionJob.main(args) == 0
    assert run_called[0] == 1


def test_main_strategies(mocker, monkeypatch, tmp_path):  # noqa pylint: disable=redefined-outer-name
    "strategies list should be respected"
    # uses the job fixture from test_reduce which reduces testcases to the string "required\n"
    report_data = {"num_reports": 0}

    class FakeReporter(BaseFakeReporter):

        def __init__(self, *args, **kwds):
            super(FakeReporter, self).__init__(*args, **kwds)
            self.report_path = "foo"

        def _submit_report(self, _report, test_cases):
            assert len(test_cases) == 1, "too many test_cases: %r" % (test_cases,)
            tc = test_cases[0]
            assert len(tc._files.required) == 1, \
                "too many test_files: %r" % (tc._files.required,)
            assert tc.landing_page == "test.html"
            assert tc._files.required[0].data == b"'xxrequired'\n"
            report_data["num_reports"] += 1
    monkeypatch.setattr(reduce, "FilesystemReporter", FakeReporter)

    (tmp_path / "binary").touch()
    exe = tmp_path / "binary"
    (tmp_path / "input").mkdir()
    inp = tmp_path / "input"
    (inp / "test_info.txt").write_text("landing page: test.html")
    (inp / "test.html").write_bytes(b"fluff\n'xxrequired'\n")
    mocker.patch("grizzly.reduce.reduce.load_target")
    args = ReducerArgs().parse_args([str(exe), str(inp), "--strategy", "line"])
    assert TestMainReductionJob.main(args) == 0
    assert report_data["num_reports"] == 1


def test_bucket_main(monkeypatch, tmp_path):  # noqa pylint: disable=redefined-outer-name
    "bucket.main iterates using crash.main"
    main_called = [0]

    class FakeCollector(object):
        serverProtocol = 'https'
        serverHost = 'mozilla.org'
        serverPort = 8000

        def get(self, url, **kwds):
            class response(object):
                @staticmethod
                def json():
                    if "crashes" in url:
                        return {
                            "results": [
                                {"testcase": None, "id": 123},
                                {"testcase": True, "id": 456},
                            ],
                            "next": None,
                            "count": 2,
                        }
                    else:
                        return {
                            "signature": '{"symptoms": []}',
                        }
            return response

    class FakeCrashReductionJob(object):

        @staticmethod
        def main(args):
            assert args.input == 456
            main_called[0] += 1
            return 0

    monkeypatch.setattr(bucket, "Collector", FakeCollector)
    monkeypatch.setattr(bucket, "CrashReductionJob", FakeCrashReductionJob)

    (tmp_path / "binary").touch()
    exe = tmp_path / "binary"
    args = ReducerFuzzManagerIDQualityArgs().parse_args([str(exe), '789'])
    assert bucket.main(args) == 0
    assert main_called[0] == 1


def test_crash_main_repro(mocker, monkeypatch, tmp_path):  # noqa pylint: disable=redefined-outer-name
    "crash.main --fuzzmanager updates quality"
    # expect Collector.patch to be called with these qualities
    expect_patch = [reporter.FuzzManagerReporter.QUAL_REPRODUCIBLE,
                    reporter.FuzzManagerReporter.QUAL_REDUCED_ORIGINAL]
    submitted = [False]

    class ReporterNoSubmit(reporter.FuzzManagerReporter):
        (tmp_path / ".fuzzmanagerconf").touch()
        FM_CONFIG = str(tmp_path / ".fuzzmanagerconf")

        def _submit_report(self, *_args, **_kwds):
            # check that the crash was already marked reproducible, but not yet marked reduced
            assert expect_patch == [reporter.FuzzManagerReporter.QUAL_REDUCED_ORIGINAL]
            submitted[0] = True

    class FakeCollector(object):
        serverProtocol = 'https'
        serverHost = 'mozilla.org'
        serverPort = 8000

        def get(self, _url, **kwds):

            class response(object):

                class headers(object):

                    @staticmethod
                    def get(value, default):
                        assert value.lower() == 'content-disposition'
                        return 'attachment; filename="test.zip"'

                @staticmethod
                def json():
                    return {
                        'testcase_quality': reporter.FuzzManagerReporter.QUAL_UNREDUCED,
                        'tool': 'test-tool'}
                content = (inp / "test.zip").read_bytes()
            return response

        def patch(self, _url, **kwds):
            data = kwds["data"]
            assert set(data.keys()) == {"testcase_quality"}
            assert expect_patch
            assert data["testcase_quality"] == expect_patch.pop(0)

    monkeypatch.setattr(reporter, "Collector", FakeCollector)
    monkeypatch.setattr(reduce, "FuzzManagerReporter", ReporterNoSubmit)
    monkeypatch.setattr(crash, "Collector", FakeCollector)

    (tmp_path / "binary").touch()
    exe = tmp_path / "binary"
    (tmp_path / "binary.fuzzmanagerconf").write_text(
        "[Main]\n"
        "platform = x86-64\n"
        "product = mozilla-central\n"
        "os = linux\n"
    )
    (tmp_path / "input").mkdir()
    inp = tmp_path / "input"
    (inp / "test_info.txt").write_text("landing page: test.html")
    (inp / "test.html").write_text("fluff\nrequired\n")
    with zipfile.ZipFile(str(inp / "test.zip"), "w") as zip_fp:
        zip_fp.write(str(inp / "test_info.txt"), "test_info.txt")
        zip_fp.write(str(inp / "test.html"), "test.html")
    args = ReducerFuzzManagerIDArgs().parse_args([str(exe), '1234', '--fuzzmanager'])
    mocker.patch("grizzly.reduce.reduce.load_target")
    assert TestMainCrashReductionJob.main(args) == 0
    assert not expect_patch
    assert submitted[0]


def test_crash_main_no_repro(mocker, monkeypatch, tmp_path):  # noqa pylint: disable=redefined-outer-name
    "crash.main --fuzzmanager updates quality"
    expect_patch = [reporter.FuzzManagerReporter.QUAL_REQUEST_SPECIFIC]

    class ReporterNoSubmit(reporter.FuzzManagerReporter):
        (tmp_path / ".fuzzmanagerconf").touch()
        FM_CONFIG = str(tmp_path / ".fuzzmanagerconf")

        def _reset(self):
            pass

        def _submit_report(self, *_args, **_kwds):
            # make sure _submit_report() is not called
            assert False

    class FakeCollector(object):
        serverProtocol = 'https'
        serverHost = 'mozilla.org'
        serverPort = 8000

        def get(self, _url, **kwds):

            class response(object):

                class headers(object):

                    @staticmethod
                    def get(value, default):
                        assert value.lower() == 'content-disposition'
                        return 'attachment; filename="test.zip"'

                @staticmethod
                def json():
                    return {
                        'testcase_quality': reporter.FuzzManagerReporter.QUAL_UNREDUCED,
                        'tool': 'test-tool'}
                content = (inp / "test.zip").read_bytes()
            return response

        def patch(self, _url, **kwds):
            data = kwds["data"]
            assert set(data.keys()) == {"testcase_quality"}
            assert expect_patch
            assert data["testcase_quality"] == expect_patch.pop(0)

    # uses the job fixture from test_reduce which reduces testcases to the string "required\n"
    monkeypatch.setattr(reporter, "Collector", FakeCollector)
    monkeypatch.setattr(reduce, "FuzzManagerReporter", ReporterNoSubmit)
    monkeypatch.setattr(crash, "Collector", FakeCollector)

    (tmp_path / "binary").touch()
    exe = tmp_path / "binary"
    (tmp_path / "binary.fuzzmanagerconf").write_text(
        "[Main]\n"
        "platform = x86-64\n"
        "product = mozilla-central\n"
        "os = linux\n"
    )
    (tmp_path / "input").mkdir()
    inp = tmp_path / "input"
    (inp / "test_info.txt").write_text("landing page: test.html")
    (inp / "test.html").write_text("fluff\n")
    with zipfile.ZipFile(str(inp / "test.zip"), "w") as zip_fp:
        zip_fp.write(str(inp / "test_info.txt"), "test_info.txt")
        zip_fp.write(str(inp / "test.html"), "test.html")
    args = ReducerFuzzManagerIDArgs().parse_args([str(exe), '1234', '--fuzzmanager'])
    mocker.patch("grizzly.reduce.reduce.load_target")
    assert TestMainCrashReductionJob.main(args) == 1
    assert not expect_patch


def test_crash_main_no_repro_specific(mocker, monkeypatch, tmp_path):  # noqa pylint: disable=redefined-outer-name
    "crash.main --fuzzmanager updates quality"
    expect_patch = [reporter.FuzzManagerReporter.QUAL_NOT_REPRODUCIBLE]

    class ReporterNoSubmit(reporter.FuzzManagerReporter):
        (tmp_path / ".fuzzmanagerconf").touch()
        FM_CONFIG = str(tmp_path / ".fuzzmanagerconf")

        def _reset(self):
            pass

        def _submit_report(self, *_args, **_kwds):
            # make sure _submit_report() is not called
            assert False

    class FakeCollector(object):
        serverProtocol = 'https'
        serverHost = 'mozilla.org'
        serverPort = 8000

        def get(self, _url, **kwds):

            class response(object):

                class headers(object):

                    @staticmethod
                    def get(value, default):
                        assert value.lower() == 'content-disposition'
                        return 'attachment; filename="test.zip"'

                @staticmethod
                def json():
                    return {
                        'testcase_quality': reporter.FuzzManagerReporter.QUAL_REQUEST_SPECIFIC,
                        'tool': 'test-tool'}
                content = (inp / "test.zip").read_bytes()
            return response

        def patch(self, _url, **kwds):
            data = kwds["data"]
            assert set(data.keys()) == {"testcase_quality"}
            assert expect_patch
            assert data["testcase_quality"] == expect_patch.pop(0)

    monkeypatch.setattr(reporter, "Collector", FakeCollector)
    monkeypatch.setattr(reduce, "FuzzManagerReporter", ReporterNoSubmit)
    monkeypatch.setattr(crash, "Collector", FakeCollector)

    (tmp_path / "binary").touch()
    exe = tmp_path / "binary"
    (tmp_path / "binary.fuzzmanagerconf").write_text(
        "[Main]\n"
        "platform = x86-64\n"
        "product = mozilla-central\n"
        "os = linux\n"
    )
    (tmp_path / "input").mkdir()
    inp = tmp_path / "input"
    (inp / "test_info.txt").write_text("landing page: test.html")
    (inp / "test.html").write_text("fluff\n")
    with zipfile.ZipFile(str(inp / "test.zip"), "w") as zip_fp:
        zip_fp.write(str(inp / "test_info.txt"), "test_info.txt")
        zip_fp.write(str(inp / "test.html"), "test.html")
    args = ReducerFuzzManagerIDArgs().parse_args([str(exe), '1234', '--fuzzmanager'])
    mocker.patch("grizzly.reduce.reduce.load_target")
    assert TestMainCrashReductionJob.main(args) == 1
    assert not expect_patch


def test_environ_and_suppressions(mocker, tmp_path):
    ""
    run_called = [0]

    class MyReductionJob(TestMainReductionJob):
        __slots__ = []

        def __init__(self, *_args, **_kwds):
            super(MyReductionJob, self).__init__(tmp_path, create_binary=False)
            assert self.target.forced_close

        def run(self, *args, **kwds):
            assert len(self._env_mod) == 2
            assert "GRZ_FORCED_CLOSE" in self._env_mod
            assert self._env_mod["GRZ_FORCED_CLOSE"] == "0"
            assert not self.target.forced_close
            assert "LSAN_OPTIONS" in self._env_mod
            assert len(re.split(r":(?![\\|/])", self._env_mod["LSAN_OPTIONS"])) == 2
            assert "detect_leaks=1" in self._env_mod["LSAN_OPTIONS"]
            assert "lsan.supp" in self._env_mod["LSAN_OPTIONS"]
            run_called[0] += 1

    exe = tmp_path / "binary"
    exe.touch()
    inp = tmp_path / "input"
    inp.mkdir()
    (inp / "env_vars.txt").write_text("LSAN_OPTIONS=detect_leaks=1\nGRZ_FORCED_CLOSE=0")
    (inp / "test_info.txt").write_text("landing page: test.html")
    (inp / "lsan.supp").write_text("foo")
    (inp / "test.html").write_text("fluff\nrequired\n")
    args = ReducerArgs().parse_args([str(exe), str(inp)])
    mocker.patch("grizzly.reduce.reduce.load_target")
    MyReductionJob.main(args)
    assert run_called[0] == 1



# coding=utf-8
"""
Sapphire unit test fixtures
"""
import hashlib
from http.client import BadStatusLine
import logging
import random
import re
import socket
import sys
import threading
import time
from urllib.error import HTTPError, URLError
from urllib.parse import quote
from urllib.request import urlopen

import pytest


LOG = logging.getLogger("sphr_test")


@pytest.fixture
def client_factory():
    """
    factory fixture to create test clients to make http requests against sapphire
    """
    clients = []

    class _SimpleClient(object):
        # pylint: disable=missing-docstring

        def __init__(self, rx_size=0x10000):
            self.thread = None
            self.rx_size = rx_size
            self._idle = threading.Event()
            self._idle.set()

        def close(self):
            if self.thread is not None:
                self.thread.join()
                self.thread = None
            self._idle.set()

        def launch(self, addr, port, files_to_serve, delay=0, in_order=False, indicate_failure=False,
                   skip_served=True, throttle=0):
            assert self._idle.is_set()
            assert self.thread is None
            self._idle.clear()
            self.thread = threading.Thread(
                target=self._handle_request,
                args=(addr, port, files_to_serve),
                kwargs={
                    "delay": delay,
                    "in_order": in_order,
                    "indicate_failure": indicate_failure,
                    "skip_served": skip_served,
                    "throttle": throttle})
            self.thread.start()

        def _handle_request(self, addr, port, files_to_request, delay=0, in_order=False,
                            indicate_failure=False, skip_served=True, throttle=0):
            assert isinstance(files_to_request, list), "files_to_request should be a list"
            if delay:
                time.sleep(delay)
            indexes = list(range(len(files_to_request)))
            if not in_order:
                random.shuffle(indexes)  # request files in random order
            for index in indexes:
                t_file = files_to_request[index]
                with t_file.lock:
                    # check if the file has been served
                    if skip_served and t_file.code is not None:
                        continue
                    # if t_file.md5_org is set to anything but None the test client will calculate
                    # the md5 hash
                    data_hash = hashlib.md5() if t_file.md5_org is not None else None
                    target_url = quote(t_file.url)
                cli = None
                try:
                    if t_file.custom_request is None:
                        cli = urlopen("http://%s:%d/%s" % (addr, port, target_url), timeout=10)
                        resp_code = cli.getcode()
                        content_type = cli.info().get("Content-Type")
                        if resp_code == 200:
                            data_length = 0
                            while True:
                                data = cli.read(self.rx_size)
                                data_length += len(data)
                                if data_hash is not None:
                                    data_hash.update(data)
                                if len(data) < self.rx_size:
                                    break
                                if throttle > 0:  # try to simulate a slow connection
                                    time.sleep(throttle)
                            if data_hash is not None:
                                data_hash = data_hash.hexdigest()
                    else:  # custom request
                        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                        try:
                            sock.connect((addr, port))
                            sock.settimeout(10)  # safety, so test doesn't hang on failure
                            sock.sendall(t_file.custom_request)
                            data = sock.recv(self.rx_size) if t_file.custom_request else b""
                        finally:
                            sock.close()
                        content_type = None
                        data_length = len(data)
                        try:
                            resp_code = int(re.match(r"HTTP/1\.\d\s(?P<code>\d+)\s",
                                                     data.decode("ascii")).group("code"))
                        except AttributeError:
                            # set code to zero to help testing
                            resp_code = 0 if indicate_failure else None

                    # update test info
                    with t_file.lock:
                        if skip_served and t_file.code is not None:
                            continue  # test has already be updated
                        t_file.requested += 1
                        t_file.code = resp_code
                        if resp_code == 200:
                            t_file.content_type = content_type
                            t_file.len_srv = data_length
                            t_file.md5_srv = data_hash

                except HTTPError as http_err:
                    with t_file.lock:
                        t_file.requested += 1
                        if not skip_served or t_file.code is None:
                            t_file.code = http_err.code
                except (BadStatusLine, socket.error, socket.timeout, URLError):
                    exc_type, exc_obj, exc_tb = sys.exc_info()
                    # set code to zero to help testing
                    with t_file.lock:
                        LOG.debug(
                            "%s - %s - line %d (processing: %s)",
                            exc_type.__name__,
                            exc_obj,
                            exc_tb.tb_lineno,
                            t_file.url)
                        if indicate_failure:
                            if not skip_served or t_file.code is None:
                                t_file.code = 0
                                break
                finally:
                    if cli is not None:
                        cli.close()
            self._idle.set()

        def wait(self, timeout=None):
            """
            Used to help prevent checking test cases before client is complete.
            """
            return self._idle.wait(timeout)

    def _get_client(*args, **kwds):
        cli = _SimpleClient(*args, **kwds)
        clients.append(cli)
        return cli

    yield _get_client

    for cli in clients:
        cli.close()


@pytest.fixture
def client(client_factory):  # pylint: disable=redefined-outer-name
    """
    create a test client to make http requests against sapphire
    """
    return client_factory()



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""
unittests for target plugin loading
"""

import pytest

from grizzly.target import available, load, Target


class _FakeTarget1(Target):  # pylint: disable=abstract-method
    pass


class _FakeTarget2(Target):  # pylint: disable=abstract-method
    pass


def test_target_load_01(mocker):
    '''If no targets are available, available() should return nothing.'''
    mocker.patch('grizzly.target.TARGETS', None)
    mocker.patch('grizzly.target.iter_entry_points', lambda _: [])
    assert not available()


def test_target_load_02(mocker):
    '''Loading targets works.'''
    mocker.patch('grizzly.target.TARGETS', None)

    class _FakeEntryPoint1(object):
        name = 'test1'

        @staticmethod
        def load():
            return _FakeTarget1

    class _FakeEntryPoint2(object):
        name = 'test2'

        @staticmethod
        def load():
            return _FakeTarget2

    mocker.patch('grizzly.target.iter_entry_points', lambda _: [_FakeEntryPoint1, _FakeEntryPoint2])
    assert set(available()) == {'test1', 'test2'}
    assert load('test1') is _FakeTarget1
    assert load('test2') is _FakeTarget2


def test_target_load_03(mocker):
    '''Non-Target will be skipped.'''
    mocker.patch('grizzly.target.TARGETS', None)

    class _FakeEntryPoint1(object):
        name = 'test1'

        @staticmethod
        def load():
            return Target

    class _FakeEntryPoint2(object):
        name = 'test2'

        @staticmethod
        def load():
            return object

    mocker.patch('grizzly.target.iter_entry_points', lambda _: [_FakeEntryPoint1, _FakeEntryPoint2])
    assert set(available()) == {'test1'}
    assert load('test1') is Target


def test_target_load_04(mocker):
    '''test load() with name collision'''
    mocker.patch('grizzly.target.TARGETS', None)

    class _FakeEntryPoint1(object):
        name = 'test'

        @staticmethod
        def load():
            return _FakeTarget1

    class _FakeEntryPoint2(object):
        name = 'test'

        @staticmethod
        def load():
            return _FakeTarget2

    mocker.patch('grizzly.target.iter_entry_points', lambda _: [_FakeEntryPoint1, _FakeEntryPoint2])
    with pytest.raises(RuntimeError, match=r"Target (.)test\1 already exists"):
        available()


def test_target_load_05(mocker):
    '''test load() with broken Target'''
    mocker.patch('grizzly.target.TARGETS', None)

    class _FakeEntryPoint1(object):
        name = 'test1'

        @staticmethod
        def load():
            return Target

    class _FakeEntryPoint2(object):
        name = 'test2'

        @staticmethod
        def load():
            raise Exception("boo!")

    mocker.patch('grizzly.target.iter_entry_points', lambda _: [_FakeEntryPoint1, _FakeEntryPoint2])
    assert set(available()) == {'test1'}
    assert load('test1') is Target



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access
from pytest import raises

from .args import GrizzlyArgs

def test_grizzly_args_01(capsys, tmp_path):
    """test GrizzlyArgs.parse_args()"""
    # test help
    with raises(SystemExit):
        GrizzlyArgs().parse_args(argv=["-h"])
    out, _ = capsys.readouterr()
    assert "For addition help check out the wiki" in out
    # test success
    fake_bin = (tmp_path / "fake.bin")
    fake_bin.touch()
    argp = GrizzlyArgs()
    argp._adapters = ["test_adapter"]
    assert argp.parse_args(argv=[str(fake_bin), "test_adapter"])

def test_grizzly_args_03(capsys):
    """test GrizzlyArgs.parse_args() handling binary"""
    # test missing required args
    with raises(SystemExit):
        GrizzlyArgs().parse_args(argv=[])
    _, err = capsys.readouterr()
    assert "the following arguments are required: binary, adapter" in err
    # test missing binary
    with raises(SystemExit):
        GrizzlyArgs().parse_args(argv=["missing_bin", "adapter"])
    _, err = capsys.readouterr()
    assert "error: file not found: 'missing_bin'" in err

def test_grizzly_args_04(capsys, tmp_path):
    """test GrizzlyArgs.parse_args() handling adapter"""
    fake_bin = (tmp_path / "fake.bin")
    fake_bin.touch()
    # no adapters
    with raises(SystemExit):
        GrizzlyArgs().parse_args(argv=[str(fake_bin), "missing"])
    _, err = capsys.readouterr()
    assert "error: Adapter 'missing' does not exist. No adapters available." in err
    # invalid adapter name
    argp = GrizzlyArgs()
    argp._adapters = ["a1", "b2"]
    with raises(SystemExit):
        argp.parse_args(argv=[str(fake_bin), "missing"])
    _, err = capsys.readouterr()
    assert "error: Adapter 'missing' does not exist. Available adapters: a1, b2" in err

# TODO: Add CommonArgs tests



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access
"""
unit tests for grizzly.Session
"""
from collections import deque

import pytest

from sapphire import Sapphire, ServerMap, SERVED_ALL, SERVED_NONE, SERVED_REQUEST, SERVED_TIMEOUT
from .common import Adapter, IOManager, Reporter, Status, TestCase
from .session import LogOutputLimiter, Session, SessionError
from .target import Target, TargetLaunchError


def test_session_01(tmp_path, mocker):
    """test Session with playback Adapter"""
    class PlaybackAdapter(Adapter):
        NAME = "playback"
        def setup(self, input_path, server_map):
            self.remaining = 5
        def generate(self, testcase, server_map):
            assert testcase.adapter_name == self.NAME
            testcase.input_fname = "file.bin"
            self.remaining -= 1
    Status.PATH = str(tmp_path)
    adapter = PlaybackAdapter()
    adapter.setup(None, None)
    fake_serv = mocker.Mock(spec=Sapphire, port=0x1337)
    fake_target = mocker.Mock(spec=Target, prefs=None)
    # set target.log_size to test warning code path
    fake_target.log_size.return_value = Session.TARGET_LOG_SIZE_WARN + 1
    with IOManager() as iomgr:
        def fake_serve_tc(tcase, **_):
            return (SERVED_ALL, [tcase.landing_page])
        fake_serv.serve_testcase = fake_serve_tc
        with Session(adapter, iomgr, None, fake_serv, fake_target) as session:
            session.run([])
            assert session.status.iteration == 5
            assert session.status.test_name == "file.bin"

def test_session_02(tmp_path, mocker):
    """test Session with basic fuzzer Adapter (w/harness)"""
    class FuzzAdapter(Adapter):
        NAME = "fuzz"
        def setup(self, input_path, server_map):
            self.enable_harness()
        def generate(self, testcase, server_map):
            assert testcase.adapter_name == self.NAME
    Status.PATH = str(tmp_path)
    adapter = FuzzAdapter()
    adapter.setup(None, None)
    fake_serv = mocker.Mock(spec=Sapphire, port=0x1337)
    fake_serv.serve_testcase.side_effect = lambda tc, **_: (SERVED_ALL, [tc.landing_page])
    fake_target = mocker.Mock(spec=Target, prefs=None, rl_reset=10)
    fake_target.log_size.return_value = 1000
    fake_target.monitor.launches = 1
    with IOManager() as iomgr:
        iomgr.harness = adapter.get_harness()
        with Session(adapter, iomgr, None, fake_serv, fake_target) as session:
            session.run([], iteration_limit=10)
            assert session.status.iteration == 10
            assert session.status.test_name is None

def test_session_03(tmp_path, mocker):
    """test Session.dump_coverage()"""
    Status.PATH = str(tmp_path)
    adapter = mocker.Mock(spec=Adapter, remaining=None)
    fake_serv = mocker.Mock(spec=Sapphire, port=0x1337)
    fake_serv.serve_testcase.side_effect = lambda tc, **_: (SERVED_ALL, [tc.landing_page])
    fake_target = mocker.Mock(spec=Target, prefs=None, rl_reset=2)
    fake_target.log_size.return_value = 1000
    fake_target.monitor.launches = 1
    # target.launch() call will be skipped
    fake_target.closed = False
    with IOManager() as iomgr:
        with Session(adapter, iomgr, None, fake_serv, fake_target, coverage=True) as session:
            session.run([], iteration_limit=2)
            assert session.status.iteration == 2
    assert fake_target.dump_coverage.call_count == 2

def test_session_04(tmp_path, mocker):
    """test Target not requesting landing page"""
    class FuzzAdapter(Adapter):
        NAME = "fuzz"
        def setup(self, input_path, server_map):
            pass
        def generate(self, testcase, server_map):
            assert testcase.adapter_name == self.NAME
    Status.PATH = str(tmp_path)
    adapter = FuzzAdapter()
    adapter.setup(None, None)
    fake_serv = mocker.Mock(spec=Sapphire, port=0x1337)
    fake_target = mocker.Mock(spec=Target, prefs=None)
    fake_target.monitor.launches = 1
    with IOManager() as iomgr:
        fake_serv.serve_testcase.return_value = (SERVED_NONE, [])
        # test error on first iteration
        with Session(adapter, iomgr, None, fake_serv, fake_target) as session:
            with pytest.raises(SessionError, match="Please check Adapter and Target"):
                session.run([], iteration_limit=10)
        # test that we continue if error happens later on
        fake_serv.serve_testcase.return_value = (SERVED_REQUEST, ["x"])
        with Session(adapter, iomgr, None, fake_serv, fake_target) as session:
            session.status.iteration = 2
            session.run([], iteration_limit=3)

def test_session_05(tmp_path, mocker):
    """test basic Session functions"""
    Status.PATH = str(tmp_path)
    fake_adapter = mocker.Mock(spec=Adapter)
    fake_adapter.TEST_DURATION = 10
    fake_adapter.remaining = None
    fake_testcase = mocker.Mock(spec=TestCase, landing_page="page.htm")
    fake_iomgr = mocker.Mock(spec=IOManager)
    fake_iomgr.server_map = ServerMap()
    fake_iomgr.create_testcase.return_value = fake_testcase
    fake_iomgr.harness = None
    fake_iomgr.tests = mocker.Mock(spec=deque)
    fake_iomgr.working_path = str(tmp_path)
    fake_serv = mocker.Mock(spec=Sapphire, port=0x1337)
    # return SERVED_TIMEOUT to test IGNORE_UNSERVED code path
    fake_serv.serve_testcase.return_value = (SERVED_TIMEOUT, [fake_testcase.landing_page])
    fake_target = mocker.Mock(spec=Target, prefs=None)
    fake_target.monitor.launches = 1
    with Session(fake_adapter, fake_iomgr, None, fake_serv, fake_target) as session:
        session.run([], iteration_limit=1)
    assert fake_adapter.setup.call_count == 0
    assert fake_adapter.pre_launch.call_count == 1
    assert fake_adapter.generate.call_count == 1
    assert fake_adapter.on_served.call_count == 0
    assert fake_adapter.on_timeout.call_count == 1
    assert fake_iomgr.create_testcase.call_count == 1
    assert fake_iomgr.tests.pop.call_count == 0
    assert fake_testcase.purge_optional.call_count == 1
    assert fake_serv.serve_testcase.call_count == 1
    assert fake_target.launch.call_count == 1
    assert fake_target.detect_failure.call_count == 1
    assert fake_target.step.call_count == 1

def test_session_06(tmp_path, mocker):
    """test Session.generate_testcase()"""
    Status.PATH = str(tmp_path)
    mocker.patch("grizzly.session.TestFile", autospec=True)
    fake_adapter = mocker.Mock(spec=Adapter)
    fake_adapter.NAME = "fake_adapter"
    fake_iomgr = mocker.Mock(spec=IOManager)
    fake_iomgr.server_map = ServerMap()
    fake_iomgr.create_testcase.return_value = mocker.Mock(spec=TestCase)
    fake_target = mocker.Mock(spec=Target, prefs="fake")
    with Session(fake_adapter, fake_iomgr, None, None, fake_target) as session:
        assert fake_adapter.generate.call_count == 0
        testcase = session.generate_testcase()
        assert fake_iomgr.create_testcase.call_count == 1
        fake_iomgr.create_testcase.assert_called_with("fake_adapter")
        assert fake_adapter.generate.call_count == 1
        fake_adapter.generate.assert_called_with(testcase, fake_iomgr.server_map)
        assert testcase.add_meta.call_count == 1

def test_session_07(tmp_path, mocker):
    """test Session.run() reporting failures"""
    Status.PATH = str(tmp_path)
    mocker.patch("grizzly.session.Report", autospec=True)
    fake_runner = mocker.patch("grizzly.session.Runner", autospec=True)
    mocker.patch("grizzly.session.TestFile", autospec=True)
    fake_adapter = mocker.Mock(spec=Adapter)
    fake_adapter.IGNORE_UNSERVED = True
    fake_adapter.TEST_DURATION = 10
    fake_adapter.remaining = None
    fake_iomgr = mocker.Mock(spec=IOManager)
    fake_iomgr.create_testcase.return_value = mocker.Mock(spec=TestCase)
    fake_iomgr.harness = None
    fake_iomgr.server_map = ServerMap()
    fake_iomgr.tests = deque()
    fake_iomgr.working_path = str(tmp_path)
    fake_reporter = mocker.Mock(spec=Reporter)
    fake_serv = mocker.Mock(spec=Sapphire, port=0x1337)
    fake_target = mocker.Mock(spec=Target, prefs="prefs.js")
    fake_target.monitor.launches = 1
    with Session(fake_adapter, fake_iomgr, fake_reporter, fake_serv, fake_target) as session:
        session.server.serve_testcase.return_value = SERVED_REQUEST
        fake_runner.return_value.result = fake_runner.return_value.FAILED
        fake_runner.return_value.served = ["/fake/file"]
        fake_runner.return_value.timeout = False
        session.run([], iteration_limit=1)
        assert fake_adapter.on_served.call_count == 1
        assert fake_adapter.on_timeout.call_count == 0
        assert fake_iomgr.purge_tests.call_count == 1
        assert fake_runner.return_value.launch.call_count == fake_iomgr.purge_tests.call_count
        assert session.status.iteration == 1
        assert session.status.results == 1
        assert session.status.ignored == 0
        assert fake_reporter.submit.call_count == 1

def test_session_08(tmp_path, mocker):
    """test Session.run() ignoring failures"""
    Status.PATH = str(tmp_path)
    fake_runner = mocker.patch("grizzly.session.Runner", autospec=True)
    mocker.patch("grizzly.session.TestFile", autospec=True)
    fake_adapter = mocker.Mock(spec=Adapter)
    fake_adapter.IGNORE_UNSERVED = True
    fake_adapter.TEST_DURATION = 10
    fake_adapter.remaining = None
    fake_iomgr = mocker.Mock(spec=IOManager)
    fake_iomgr.create_testcase.return_value = mocker.Mock(spec=TestCase)
    fake_iomgr.harness = None
    fake_iomgr.server_map = ServerMap()
    fake_iomgr.tests = mocker.Mock(spec=deque)
    fake_iomgr.tests.pop.return_value = mocker.Mock(spec=TestCase)
    fake_serv = mocker.Mock(spec=Sapphire, port=0x1337)
    fake_target = mocker.Mock(spec=Target, prefs="prefs.js")
    fake_target.monitor.launches = 1
    # ignored results should not be reported so raise AssertionError if report_result is called
    mocker.patch.object(Session, 'report_result', side_effect=AssertionError)
    with Session(fake_adapter, fake_iomgr, None, fake_serv, fake_target) as session:
        session.server.serve_testcase.return_value = SERVED_REQUEST
        fake_runner.return_value.result = fake_runner.return_value.IGNORED
        fake_runner.return_value.served = []
        fake_runner.return_value.timeout = False
        session.run([], iteration_limit=1)
        assert fake_adapter.on_served.call_count == 1
        assert fake_adapter.on_timeout.call_count == 0
        assert fake_iomgr.purge_tests.call_count == 1
        assert fake_iomgr.tests.pop.call_count == 1
        assert fake_runner.return_value.launch.call_count == fake_iomgr.purge_tests.call_count
        assert session.status.iteration == 1
        assert session.status.results == 0
        assert session.status.ignored == 1

def test_session_09(tmp_path, mocker):
    """test Session.run() handle TargetLaunchError"""
    Status.PATH = str(tmp_path)
    mocker.patch("grizzly.session.Report", autospec=True)
    fake_runner = mocker.patch("grizzly.session.Runner", autospec=True)
    fake_runner.return_value.launch.side_effect = TargetLaunchError
    mocker.patch("grizzly.session.TestFile", autospec=True)
    fake_adapter = mocker.Mock(spec=Adapter)
    fake_iomgr = mocker.Mock(spec=IOManager)
    fake_iomgr.harness = None
    fake_iomgr.input_files = []
    fake_iomgr.server_map = ServerMap()
    fake_iomgr.tests = deque()
    fake_iomgr.working_path = str(tmp_path)
    fake_reporter = mocker.Mock(spec=Reporter)
    fake_serv = mocker.Mock(spec=Sapphire, port=0x1337)
    fake_target = mocker.Mock(spec=Target)
    fake_target.monitor.launches = 1
    with Session(fake_adapter, fake_iomgr, fake_reporter, fake_serv, fake_target) as session:
        with pytest.raises(TargetLaunchError):
            session.run([], iteration_limit=1)
        assert session.status.iteration == 1
        assert session.status.results == 1
        assert session.status.ignored == 0
    assert fake_reporter.submit.call_count == 1

def test_session_10(tmp_path, mocker):
    """test Session.report_result()"""
    tmpd = tmp_path / "fake_temp_path"
    tmpd.mkdir()
    (tmpd / "log_stderr.txt").write_bytes(b"STDERR log\n")
    (tmpd / "log_stdout.txt").write_bytes(b"STDOUT log\n")
    with (tmpd / "log_asan_blah.txt").open("wb") as log_fp:
        log_fp.write(b"==1==ERROR: AddressSanitizer: ")
        log_fp.write(b"SEGV on unknown address 0x0 (pc 0x0 bp 0x0 sp 0x0 T0)\n")
        log_fp.write(b"    #0 0xbad000 in foo /file1.c:123:234\n")
        log_fp.write(b"    #1 0x1337dd in bar /file2.c:1806:19\n")
    fake_report = mocker.patch("grizzly.session.Report", autospec=True)
    mocker.patch("grizzly.session.mkdtemp", autospec=True, return_value=str(tmpd))
    Status.PATH = str(tmp_path)
    fake_iomgr = mocker.Mock(spec=IOManager)
    fake_iomgr.tests = deque()
    fake_iomgr.working_path = str(tmp_path)
    fake_reporter = mocker.Mock(spec=Reporter)
    fake_target = mocker.Mock(spec=Target, binary="bin")
    with Session(None, fake_iomgr, fake_reporter, None, fake_target) as session:
        session.report_result()
    assert fake_target.save_logs.call_count == 1
    fake_target.save_logs.assert_called_with(str(tmpd))
    assert fake_report.from_path.return_value.crash_info.call_count == 1
    fake_report.from_path.return_value.crash_info.assert_called_with("bin")
    assert fake_reporter.submit.call_count == 1
    assert not tmpd.is_dir()

def test_log_output_limiter_01(mocker):
    """test LogOutputLimiter.ready() not ready"""
    fake_time = mocker.patch("grizzly.session.time", autospec=True)
    fake_time.return_value = 1.0
    lol = LogOutputLimiter(delay=10, delta_multiplier=2)
    assert lol._delay == 10
    assert lol._iterations == 1
    assert lol._launches == 1
    assert lol._multiplier == 2
    assert lol._time == 1.0
    assert not lol._verbose
    fake_time.return_value = 1.1
    assert not lol.ready(0, 0)
    assert lol._iterations == 1
    assert lol._launches == 1
    assert lol._time == 1.0
    lol._verbose = True
    assert lol.ready(0, 0)

def test_log_output_limiter_02(mocker):
    """test LogOutputLimiter.ready() due to iterations"""
    fake_time = mocker.patch("grizzly.session.time", autospec=True)
    fake_time.return_value = 1.0
    lol = LogOutputLimiter(delay=10, delta_multiplier=2)
    fake_time.return_value = 1.1
    lol._launches = 2
    assert lol.ready(1, 1)
    assert lol._iterations == 2
    assert lol._launches == 2
    assert lol._time == 1.1

def test_log_output_limiter_03(mocker):
    """test LogOutputLimiter.ready() due to launches"""
    fake_time = mocker.patch("grizzly.session.time", autospec=True)
    fake_time.return_value = 1.0
    lol = LogOutputLimiter(delay=10, delta_multiplier=2)
    lol._iterations = 4
    assert lol.ready(3, 1)
    assert lol._launches == 2
    assert lol._iterations == 4
    assert lol._time == 1.0

def test_log_output_limiter_04(mocker):
    """test LogOutputLimiter.ready() due to time"""
    fake_time = mocker.patch("grizzly.session.time", autospec=True)
    fake_time.return_value = 1.0
    lol = LogOutputLimiter(delay=1, delta_multiplier=2)
    lol._iterations = 4
    fake_time.return_value = 2.0
    assert lol.ready(3, 0)
    assert lol._iterations == 4
    assert lol._launches == 1
    assert lol._time == 2.0



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from os.path import isfile, isdir, join as pathjoin

from ..args import CommonArgs


class ReplayArgs(CommonArgs):

    def __init__(self):
        super(ReplayArgs, self).__init__()
        self.parser.add_argument(
            "input",
            help="Directory containing test case data or file to use as a test case." \
                "When using a directory it must contain a 'test_info.json' file.")

        replay_args = self.parser.add_argument_group("Replay Arguments")
        replay_args.add_argument(
            "--any-crash", action="store_true",
            help="Any crash is interesting, not only crashes which match the original signature.")
        replay_args.add_argument(
            "--idle-threshold", type=int, default=25,
            help="CPU usage threshold to mark the process as idle (default: %(default)s)")
        replay_args.add_argument(
            "--idle-timeout", type=int, default=60,
            help="Number of seconds to wait before polling testcase for idle (default: %(default)s)")
        replay_args.add_argument(
            "-l", "--logs",
            help="Location to save logs. If the path exists it must be empty, if it " \
                "does not exist it will be created.")
        replay_args.add_argument(
            "--min-crashes", type=int, default=1,
            help="Require the testcase to crash n times before accepting the result." \
                " Helpful for intermittent testcases (default: %(default)sx)")
        replay_args.add_argument(
            "--no-harness", action="store_true",
            help="Don't use the harness for redirection")
        replay_args.add_argument(
            "--repeat", type=int, default=1,
            help="Run the testcase n times." \
                " Helpful for intermittent testcases (default: %(default)sx)")
        replay_args.add_argument(
            "--sig",
            help="Signature (JSON) file to match.")

        self.launcher_grp.add_argument(
            "--rr", action="store_true",
            help="Use RR (Linux only)")

        self.reporter_grp.add_argument(
            "--include-test", action="store_true",
            help="Include the testcase when reporting results.")

        self.parser.epilog = "For addition help check out the wiki:" \
            " https://github.com/MozillaSecurity/grizzly/wiki"

    def sanity_check(self, args):
        super(ReplayArgs, self).sanity_check(args)

        if "input" not in self._sanity_skip and isdir(args.input):
            if not isfile(pathjoin(args.input, "test_info.json")):
                self.parser.error("Test case folder must contain 'test_info.json'")

        if args.any_crash and args.sig is not None:
            self.parser.error("signature is ignored when running with '--any-crash'")

        if args.min_crashes < 1:
            self.parser.error("'--min-crashes' value must be positive")

        if args.repeat < 1:
            self.parser.error("'--repeat' value must be positive")

        if args.sig is not None and not isfile(args.sig):
            self.parser.error("signature file not found: %r" % (args.sig,))



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""test Grizzly status"""
# pylint: disable=protected-access

from multiprocessing import Event, Process
from os import remove, stat
from os.path import isfile
from time import sleep, time

from .status import ReducerStats, Status


def test_status_01(tmp_path):
    """test Status.start()"""
    working_path = tmp_path / "grzstatus"
    Status.PATH = str(working_path)
    status = Status.start()
    assert status is not None
    assert working_path.is_dir()
    assert isfile(status.data_file)
    assert stat(status.data_file).st_size > 0
    assert status.start_time > 0
    assert status.timestamp >= status.start_time
    assert int(status.duration) == 0
    assert status.ignored == 0
    assert status.iteration == 0
    assert status.log_size == 0
    assert status.rate == 0
    assert status.results == 0

def test_status_02(tmp_path):
    """test Status.cleanup()"""
    Status.PATH = str(tmp_path / "grzstatus")
    status = Status.start()
    dfile = status.data_file
    status.cleanup()
    assert status.data_file is None
    assert not isfile(dfile)
    # call 2nd time
    status.cleanup()
    # missing data file
    status = Status.start()
    remove(status.data_file)
    status.cleanup()

def test_status_03(tmp_path):
    """test Status.report()"""
    Status.PATH = str(tmp_path / "grzstatus")
    status = Status.start()
    # try to report before REPORT_FREQ elapses
    assert not status.report()
    # REPORT_FREQ elapses
    status.timestamp = 0
    assert status.report()
    assert status.timestamp > 0
    # force report
    future = int(time()) + 1000
    status.timestamp = future
    assert status.report(force=True)
    assert status.timestamp < future

def test_status_04(tmp_path):
    """test Status.load() failure paths"""
    working_path = tmp_path / "grzstatus"
    working_path.mkdir()
    Status.PATH = str(working_path)
    # load no db
    assert Status.load(str(tmp_path / "missing")) is None
    # load empty
    bad = (working_path / "bad.json")
    bad.touch()
    assert Status.load(str(bad)) is None
    # load invalid/incomplete json
    bad.write_bytes(b"{}")
    assert Status.load(str(bad)) is None

def test_status_05(tmp_path):
    """test Status.load()"""
    Status.PATH = str(tmp_path / "grzstatus")
    # create simple entry
    status = Status.start()
    loaded = Status.load(status.data_file)
    assert loaded is not None
    assert status.start_time == loaded.start_time
    assert status.timestamp == loaded.timestamp
    assert status.duration == loaded.duration
    assert status.ignored == loaded.ignored
    assert status.iteration == loaded.iteration
    assert status.log_size == loaded.log_size
    assert status.results == loaded.results

def test_status_06(tmp_path):
    """test Status.loadall()"""
    working_path = tmp_path / "grzstatus"
    Status.PATH = str(working_path)
    # missing path
    assert not any(Status.loadall())
    # no status data
    working_path.mkdir()
    assert not any(Status.loadall())
    # add more entries
    for _ in range(5):
        Status.start()
    (working_path / "empty.json").touch()
    assert len(tuple(Status.loadall())) == 5

def test_status_07(tmp_path):
    """test Status.duration and Status.rate calculations"""
    Status.PATH = str(tmp_path / "grzstatus")
    status = Status.start()
    status.start_time = 1
    status.timestamp = 2
    status.iteration = 0
    assert status.duration == 1
    assert status.rate == 0
    status.iteration = 1
    assert status.rate == 1
    status.timestamp += 1
    assert status.rate == 0.5

def _client_writer(done, working_path):
    """Used by test_status_08"""
    # NOTE: this must be at the top level to work on Windows
    Status.PATH = working_path
    status = Status.start()
    try:
        while not done.is_set():
            status.iteration += 1
            status.report(force=True)
            sleep(0.01)
    finally:
        status.cleanup()

def test_status_08(tmp_path):
    """test Status.loadall() with multiple active reporters"""
    Status.PATH = str(tmp_path / "grzstatus")
    best_rate = 0
    done = Event()
    procs = list()
    try:
        for _ in range(5):
            procs.append(Process(target=_client_writer, args=(done, Status.PATH)))
            procs[-1].start()
        deadline = time() + 60
        while len(tuple(Status.loadall())) < len(procs):
            sleep(0.1)
            assert time() < deadline, "timeout waiting for processes to launch!"
        for _ in range(20):
            for obj in Status.loadall():
                if obj.rate > best_rate:
                    best_rate = obj.rate
    finally:
        done.set()
        for proc in procs:
            if proc.pid is not None:
                proc.join()
    assert best_rate > 0
    assert not any(Status.loadall())

def test_reducer_stats_01(tmp_path):
    """test ReducerStats() empty"""
    ReducerStats.PATH = str(tmp_path)
    with ReducerStats() as stats:
        assert stats.error == 0
        assert stats.failed == 0
        assert stats.passed == 0
        stats_file = stats._file
        assert not isfile(stats_file)
    assert isfile(stats_file)

def test_reducer_stats_02(tmp_path):
    """test ReducerStats() simple"""
    ReducerStats.PATH = str(tmp_path)
    with ReducerStats() as stats:
        stats.error += 1
        stats.failed += 1
        stats.passed += 1
    with ReducerStats() as stats:
        assert stats.error == 1
        assert stats.failed == 1
        assert stats.passed == 1

def test_reducer_stats_03(tmp_path):
    """test ReducerStats() empty/incomplete/invalid data file"""
    ReducerStats.PATH = str(tmp_path)
    stats_file = tmp_path / ReducerStats.FILE
    # missing file
    with ReducerStats() as stats:
        stats.passed += 1
    # invalid empty file
    stats_file.write_bytes(b"")
    with ReducerStats() as stats:
        assert stats.passed == 0
    # incomplete file
    stats_file.write_bytes(b"{}")
    with ReducerStats() as stats:
        assert stats.passed == 0

def _reducer_client(working_path, limit, unrestrict):
    """Used by test_reducer_stats_04"""
    # NOTE: this must be at the top level to work on Windows
    ReducerStats.PATH = working_path
    for _ in range(50):
        with ReducerStats() as stats:
            stats.passed += 1
            if stats.passed == limit:
                unrestrict.set()
        unrestrict.wait(timeout=60)

def test_reducer_stats_04(tmp_path):
    """test ReducerStats() with multiple processes"""
    ReducerStats.PATH = str(tmp_path)
    procs = list()
    unrestrict = Event()  # used to sync client procs
    try:
        proc_count = 5
        for _ in range(proc_count):
            procs.append(Process(
                target=_reducer_client, args=(ReducerStats.PATH, proc_count, unrestrict)))
            procs[-1].start()
    finally:
        unrestrict.set()
        for proc in procs:
            if proc.pid is not None:
                proc.join()
    with ReducerStats() as stats:
        assert stats.passed == 250



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""
Sapphire HTTP server worker
"""

import mimetypes
from logging import getLogger
import os
import re
import socket
import sys
import threading
import time
from urllib.parse import unquote_plus

from .server_map import Resource

__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]

LOG = getLogger("sphr_worker")


class SapphireWorkerError(Exception):
    """Raised by SapphireWorker"""


class SapphireWorker(object):
    DEFAULT_REQUEST_LIMIT = 0x1000  # 4KB
    DEFAULT_TX_SIZE = 0x10000  # 64KB
    REQ_PATTERN = re.compile(b"^GET\\s/(?P<request>\\S*)\\sHTTP/1")

    __slots__ = ("_conn", "_thread")

    def __init__(self, conn, thread):
        self._conn = conn
        self._thread = thread

    @staticmethod
    def _200_header(c_length, c_type, encoding="ascii"):
        data = "HTTP/1.1 200 OK\r\n" \
               "Cache-Control: max-age=0, no-cache\r\n" \
               "Content-Length: %d\r\n" \
               "Content-Type: %s\r\n" \
               "Connection: close\r\n\r\n" % (c_length, c_type)
        return data.encode(encoding)

    @staticmethod
    def _307_redirect(redirct_to, encoding="ascii"):
        data = "HTTP/1.1 307 Temporary Redirect\r\n" \
               "Location: %s\r\n" \
               "Connection: close\r\n\r\n" % (redirct_to,)
        return data.encode(encoding)

    @staticmethod
    def _4xx_page(code, hdr_msg, close=-1, encoding="ascii"):
        if close < 0:
            content = "<h3>%d!</h3>" % (code,)
        else:
            content = "<script>window.setTimeout(window.close, %d)</script>\n" \
                      "<body style=\"background-color:#ffffe0\">\n" \
                      "<h3>%d! - Calling window.close() in %d seconds</h3>\n" \
                      "</body>\n" % (close * 1000, code, close)
        data = "HTTP/1.1 %d %s\r\n" \
               "Content-Length: %d\r\n" \
               "Content-Type: text/html\r\n" \
               "Connection: close\r\n\r\n%s" % (code, hdr_msg, len(content), content)
        return data.encode(encoding)

    def close(self):
        if not self.done:
            LOG.debug("closing socket while thread is running!")
        self._conn.close()
        self.join(timeout=60)
        if self._thread is not None and self._thread.is_alive():
            # this is here to catch unexpected hangs
            raise SapphireWorkerError("Worker thread failed to join!")

    @property
    def done(self):
        if self._thread is not None and not self._thread.is_alive():
            self.join()
            self._thread = None
        return self._thread is None

    @classmethod
    def handle_request(cls, conn, serv_job):
        finish_job = False  # call finish() on return
        try:
            # receive all the incoming data
            raw_request = conn.recv(cls.DEFAULT_REQUEST_LIMIT)
            if not raw_request:
                LOG.debug("raw_request was empty")
                serv_job.accepting.set()
                return

            request = cls.REQ_PATTERN.match(raw_request)
            if request is None:
                serv_job.accepting.set()
                conn.sendall(cls._4xx_page(400, "Bad Request", serv_job.auto_close))
                LOG.debug("400 request length %d (%d to go)", len(raw_request), serv_job.pending)
                return

            request = unquote_plus(request.group("request").decode("ascii"))
            LOG.debug("check_request(%r)", request)
            resource = serv_job.check_request(request)
            if resource is None:
                LOG.debug("resource is None")  # 404
            elif resource.type in (Resource.URL_FILE, Resource.URL_INCLUDE):
                finish_job = serv_job.remove_pending(resource.target)
            elif resource.type == Resource.URL_REDIRECT:
                finish_job = serv_job.remove_pending(request)

            if finish_job and serv_job.forever:
                LOG.debug("serv_job.forever is set, resetting finish_job")
                finish_job = False

            if not finish_job:
                serv_job.accepting.set()
            else:
                LOG.debug("expecting to finish")

            if resource is None:
                conn.sendall(cls._4xx_page(404, "Not Found", serv_job.auto_close))
                LOG.debug("404 %r (%d to go)", request, serv_job.pending)
                return
            if resource.type in (Resource.URL_FILE, Resource.URL_INCLUDE):
                LOG.debug("target %r", resource.target)
                if not os.path.isfile(resource.target):
                    conn.sendall(cls._4xx_page(404, "Not Found", serv_job.auto_close))
                    LOG.debug("404 %r (%d to go)", request, serv_job.pending)
                    return
                if serv_job.is_forbidden(resource.target):
                    # NOTE: this does info leak if files exist on disk.
                    # We could replace 403 with 404 if it turns out we care but this
                    # is meant to run locally and only be accessible from localhost
                    conn.sendall(cls._4xx_page(403, "Forbidden", serv_job.auto_close))
                    LOG.debug("403 %r (%d to go)", request, serv_job.pending)
                    return
            elif resource.type == Resource.URL_REDIRECT:
                conn.sendall(cls._307_redirect(resource.target))
                LOG.debug(
                    "307 %r -> %r (%d to go)",
                    request,
                    resource.target,
                    serv_job.pending)
                return
            elif resource.type == Resource.URL_DYNAMIC:
                data = resource.target()
                if not isinstance(data, bytes):
                    LOG.debug("dynamic request: %r", request)
                    raise TypeError("dynamic request callback must return 'bytes'")
                conn.sendall(cls._200_header(len(data), resource.mime))
                conn.sendall(data)
                LOG.debug("200 %r (dynamic request)", request)
                return
            else:
                raise SapphireWorkerError("Unknown resource type %r" % (resource.type,))

            # at this point we know "resource.target" maps to a file on disk
            # default to "application/octet-stream"
            c_type = mimetypes.guess_type(resource.target)[0] or "application/octet-stream"
            # serve the file
            data_size = os.stat(resource.target).st_size
            LOG.debug("sending file: %s bytes", format(data_size, ","))
            with open(resource.target, "rb") as in_fp:
                conn.sendall(cls._200_header(data_size, c_type))
                offset = 0
                while offset < data_size:
                    conn.sendall(in_fp.read(cls.DEFAULT_TX_SIZE))
                    offset = in_fp.tell()
            LOG.debug("200 %r (%d to go)", resource.target, serv_job.pending)
            serv_job.increment_served(resource.target)

        except (socket.timeout, socket.error):
            exc_type, exc_obj, exc_tb = sys.exc_info()
            LOG.debug("%s: %r (line %d)", exc_type.__name__, exc_obj, exc_tb.tb_lineno)
            if not finish_job:
                serv_job.accepting.set()

        except Exception:  # pylint: disable=broad-except
            # set finish_job to abort immediately
            finish_job = True
            if serv_job.exceptions.empty():
                serv_job.exceptions.put(sys.exc_info())

        finally:
            conn.close()
            if finish_job:
                serv_job.finish()
            serv_job.worker_complete.set()

    def join(self, timeout=None):
        if self._thread is not None:
            self._thread.join(timeout=timeout)
            if not self._thread.is_alive():
                self._thread = None

    @classmethod
    def launch(cls, listen_sock, job):
        assert job.accepting.is_set()
        conn = None
        try:
            conn, _ = listen_sock.accept()
            conn.settimeout(None)
            # create a worker thread to handle client request
            w_thread = threading.Thread(target=cls.handle_request, args=(conn, job))
            job.accepting.clear()
            w_thread.start()
            return cls(conn, w_thread)
        except (socket.error, socket.timeout):
            if conn is not None:  # pragma: no cover
                conn.close()
        except threading.ThreadError:
            if conn is not None:  # pragma: no cover
                conn.close()
            # reset accepting status
            job.accepting.set()
            LOG.warning("ThreadError (worker), threads: %d", threading.active_count())
            # wait for system resources to free up
            time.sleep(0.1)
        return None



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from .adapter import Adapter, AdapterError
from .iomanager import IOManager, ServerMap
from .reporter import FilesystemReporter, FuzzManagerReporter, Report, Reporter, S3FuzzManagerReporter
from .runner import Runner
from .status import ReducerStats, Status
from .storage import TestCaseLoadFailure, TestCase, TestFile, TestFileExists
from .utils import grz_tmp


__all__ = (
    "Adapter", "AdapterError", "FilesystemReporter", "FuzzManagerReporter", "grz_tmp", "IOManager",
    "ReducerStats", "Report", "Reporter", "Runner", "S3FuzzManagerReporter", "ServerMap", "Status",
    "TestCase", "TestCaseLoadFailure", "TestFile", "TestFileExists")
__author__ = "Jesse Schwartzentruber"
__credits__ = ["Jesse Schwartzentruber", "Tyson Smith"]



# coding=utf-8
"""
Loki fuzzing library
"""
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from .loki import Loki  # noqa

__all__ = ("Loki")
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""
Given a build and testcase, try to reproduce it using a set of strategies.
"""
from __future__ import absolute_import
import glob
import hashlib
import io
import json
import logging
import os
import re
import shutil
import tempfile
import time
import zipfile
import zlib

import lithium
import sapphire
from FTB.Signatures.CrashInfo import CrashSignature

from . import strategies as strategies_module, testcase_contents
from .exceptions import CorruptTestcaseError, NoTestcaseError, ReducerError
from ..common.reporter import FilesystemReporter, FuzzManagerReporter, Report
from ..common.runner import Runner
from ..common.status import ReducerStats, Status
from ..common.storage import TestCase, TestFile
from ..common.utils import grz_tmp
from ..main import configure_logging
from ..session import Session
from ..target import load as load_target, sanitizer_opts, TargetLaunchError, \
    TargetLaunchTimeout


__author__ = "Jesse Schwartzentruber"
__credits__ = ["Tyson Smith", "Jesse Schwartzentruber", "Jason Kratzer"]


LOG = logging.getLogger("grizzly.reduce")


class LithiumInterestingProxy(object):
    """Proxy to use a ReductionJob object as a Lithium interestingness script object.
    """
    __slots__ = ['_job']

    def __init__(self, job):
        self._job = job

    def init(self, _args):
        """Lithium initialization entrypoint.

        Do any per-reduction loop setup needed.

        Args:
            _args (unused): Command line arguments from Lithium (N/A)

        Returns:
            None
        """
        self._job.lithium_init()

    def interesting(self, _args, temp_prefix):
        """Lithium main iteration entrypoint.

        This should try the reduction and return True or False based on whether the reduction was
        good or bad.

        Args:
            _args (unused): Command line arguments from Lithium (N/A)
            temp_prefix (str): A unique prefix for any files written during this iteration.

        Returns:
            bool: True if reduced testcase is still interesting.
        """
        return self._job.lithium_interesting(temp_prefix)

    def cleanup(self, _args):
        """Lithium cleanup entrypoint.

        Do any per-reduction loop cleanup needed.

        Args:
            _args (unused): Command line arguments from Lithium (N/A)

        Returns:
            None
        """
        self._job.lithium_cleanup()


class IterationParamsProxy(object):
    __slots__ = ['_fixed_timeout', '_job', '_use_result_cache']

    def __init__(self, job):
        self._job = job
        self._use_result_cache = None
        self._fixed_timeout = None

    def __enter__(self):
        # disable result cache setting
        self._use_result_cache = self._job._use_result_cache
        self._job._use_result_cache = False

        # do not update the iteration timeout during analysis
        self._fixed_timeout = self._job._fixed_timeout
        self._job._fixed_timeout = True

        return self

    def __exit__(self, *_args):
        # restore saved values
        self._job._use_result_cache = self._use_result_cache
        self._job._fixed_timeout = self._fixed_timeout

    @property
    def force_no_harness(self):
        return self._job._force_no_harness

    @property
    def min_crashes(self):
        return self._job._min_crashes

    @min_crashes.setter
    def min_crashes(self, value):
        self._job._min_crashes = value

    @property
    def no_harness(self):
        return self._job._no_harness

    @no_harness.setter
    def no_harness(self, value):
        self._job._no_harness = value

    @property
    def relaunch(self):
        return self._job._target.rl_reset

    @relaunch.setter
    def relaunch(self, value):
        self._job._target.rl_reset = min(self._job._original_relaunch, value)

    @property
    def repeat(self):
        return self._job._repeat

    @repeat.setter
    def repeat(self, value):
        self._job._repeat = value

    def commit(self):
        # close target so new parameters take effect
        self._job.close_target()


class TimeoutsUpdateProxy(object):
    __slots__ = ['_job']

    def __init__(self, job):
        self._job = job

    @property
    def idle(self):
        return self._job._idle_timeout

    @idle.setter
    def idle(self, value):
        self._job._idle_timeout = value

    @property
    def iteration(self):
        return self._job._iter_timeout

    @iteration.setter
    def iteration(self, value):
        self._job._iter_timeout = value


class RunState(object):
    __slots__ = ['files_to_reduce', 'original_size']

    def __init__(self, files_to_reduce):
        self.files_to_reduce = files_to_reduce
        self.original_size = -1

    def total_size(self):
        return sum(os.stat(fn).st_size for fn in self.files_to_reduce)


class TestcaseUpdateProxy(object):
    __slots__ = ['_job', '_run_state']

    def __init__(self, job, run_state):
        self._job = job
        self._run_state = run_state

    @property
    def cache_iter_harness_created(self):
        return self._job._cache_iter_harness_created

    @property
    def root(self):
        return self._job._tcroot

    @root.setter
    def root(self, value):
        self._job._tcroot = value

    @property
    def entry(self):
        return self._job._testcase

    @entry.setter
    def entry(self, value):
        self._job._testcase = value

    @property
    def landing_page(self):
        return self._job.landing_page

    @landing_page.setter
    def landing_page(self, value):
        self._job.landing_page = value

    @property
    def files_to_reduce(self):
        return self._run_state.files_to_reduce

    @property
    def original_size(self):
        return self._run_state.original_size

    @original_size.setter
    def original_size(self, value):
        self._run_state.original_size = value

    def total_size(self):
        return self._run_state.total_size()


class ReductionJob(object):
    LOGGERS_TO_WATCH = ("ffpuppet", "grizzly", "lithium", "sapphire")
    DEFAULT_STRATEGIES = ("line", "cssbeautify", "jsbeautify", "collapsebraces", "jschar")
    __slots__ = [
        '_any_crash', '_best_testcase', '_cache_iter_harness_created', '_env_mod',
        '_fixed_timeout', '_force_no_harness', '_idle_threshold', '_idle_timeout', '_ignore',
        '_input_fname', '_interesting_prefix', '_iter_timeout', '_landing_page', '_log_handler',
        '_min_crashes', '_no_harness', '_orig_sig', '_original_relaunch', '_other_crashes',
        '_reduce_file', '_repeat', '_reporter', '_result_cache', '_result_code', '_server', '_server_map',
        '_signature', '_skip', '_skip_analysis', '_skipped', '_status', '_target', '_tcroot', '_testcase',
        '_tmpdir', '_use_result_cache',
    ]

    def __init__(self, ignore, target, iter_timeout, no_harness, any_crash, skip, min_crashes,
                 repeat, idle_threshold, idle_timeout, status, testcase_cache=True, skip_analysis=False):
        """Use lithium to reduce a testcase.

        Args:
            target (grizzly.target.Target): Target object to use for reduction.
        """
        self._any_crash = any_crash
        self._best_testcase = None
        self._cache_iter_harness_created = None
        self._env_mod = None  # environment if specified in the testcase
        self._fixed_timeout = False  # if True iter_timeout will not be changed
        self._force_no_harness = no_harness
        self._idle_threshold = idle_threshold
        self._idle_timeout = idle_timeout
        self._ignore = ignore  # things to ignore
        self._input_fname = None
        self._interesting_prefix = None
        self._iter_timeout = iter_timeout
        self._landing_page = None  # the file to point the target at
        self._min_crashes = min_crashes
        self._no_harness = no_harness
        self._orig_sig = None  # signature to reduce to (if specified)
        self._original_relaunch = target.rl_reset
        self._other_crashes = {}
        self._reduce_file = None  # the file to reduce
        self._repeat = repeat
        self._reporter = None
        self._result_cache = {}
        self._result_code = None
        self._server = None  # a server to serve with
        self._server_map = sapphire.ServerMap()  # manage dynamic requests, includes and redirects
        self._signature = None
        self._skip = skip
        self._skip_analysis = skip_analysis
        self._skipped = None
        self._status = status  # ReduceStatus to track progress
        self._target = target  # a Puppet to run with
        self._testcase = None
        # testcase cache remembers if we have seen this reduce_file before and if so return the same
        # interesting result
        self._use_result_cache = testcase_cache
        self._tmpdir = tempfile.mkdtemp(prefix="grzreduce", dir=grz_tmp("reduce"))
        self._tcroot = os.path.join(self._tmpdir, "tc")
        self._log_handler = self._start_log_capture()
        if not self._skip_analysis:
            # see if any of the args tweaked by analysis were overridden
            # --relaunch is regarded as a maximum, so overriding the default is not a deal-breaker for this
            if self._min_crashes != 1:
                LOG.warning("--min-crashes=%d was given, skipping analysis", self._min_crashes)
                self._skip_analysis = True
            elif self._repeat != 1:
                LOG.warning("--repeat=%d was given, skipping analysis", self._repeat)
                self._skip_analysis = True

    @property
    def landing_page(self):
        return os.path.basename(self._landing_page)

    @landing_page.setter
    def landing_page(self, value):
        # this looks pointless, but it isn't since it affects both landing_page and wwwdir getters
        self._landing_page = value

    @property
    def reduce_file(self):
        return self._reduce_file

    @reduce_file.setter
    def reduce_file(self, value):
        self._reduce_file = value
        # landing page should default to same value as reduce file
        if self._landing_page is None:
            self._landing_page = value

    @property
    def result_code(self):
        return self._result_code

    @property
    def server(self):
        return self._server

    @property
    def target(self):
        return self._target

    @property
    def wwwdir(self):
        return os.path.dirname(os.path.realpath(self._landing_page))

    def timeouts_proxy(self):
        """Return a proxy for modifying the job timeouts.

        Returns:
            (object): an object used to modify the timeouts for this job
                      attributes:
                          - iteration (iteration timeout)
                          - idle (idle timeout)
        """
        return TimeoutsUpdateProxy(self)

    def testcase_proxy(self, run_state):
        """Return a proxy for modifying the testcase.

        Returns:
            (object): an object used to modify the testcase for this job
                      attributes:
                          - iteration (iteration timeout)
                          - idle (idle timeout)
        """
        return TestcaseUpdateProxy(self, run_state)

    def analysis_mode(self, min_crashes=1, relaunch=1, repeat=1):
        """Set parameters for testcase analysis. This has side-effects besides being a proxy:

            - min_crashes/repeat/relaunch are preset according to the function parameters
            - result cache is disabled (so every iteration runs fully) if used as a with-statement context
            - times are not dynamically adjusted if used as a with-statement context

        Args:
            min_crashes (int): How many crashes are needed for a success.
            relaunch (int): How many iterations between relaunch.
            repeat (int): How many times to repeat the testcase per iteration.

        Returns:
            (context manager): an object that can be used to set new parameters
                               as a result of analysis:
                               attributes:
                                   - min_crashes
                                   - no_harness
                                   - relaunch
                                   - repeat
        """
        # pylint: disable=no-self-argument,no-self-use,protected-access

        proxy = IterationParamsProxy(self)

        # Set parameters for analysis
        proxy.min_crashes = min_crashes
        proxy.repeat = repeat
        proxy.relaunch = relaunch

        return proxy

    def close_target(self):
        if not self._target.closed:
            self._target.close()

    def lithium_init(self):
        """Lithium initialization entrypoint. Do any per-reduction loop setup needed.

        Args:
            None

        Returns:
            None
        """
        self._skipped = None
        self._best_testcase = None
        self._result_cache = {}

    def lithium_interesting(self, temp_prefix):
        """Lithium main iteration entrypoint.

        This should try the reduction and return True or False based on whether the reduction was
        good or bad.  This is subject to a number of options (skip, repeat, cache) and so may
        result in 0 or more actual runs of the target.

        Args:
            temp_prefix (str): A unique prefix for any files written during this iteration.

        Returns:
            bool: True if reduced testcase is still interesting.
        """
        # ensure the target is closed so "repeat" and "relaunch" never get out of sync
        self.close_target()
        if self._skip:
            if self._skipped is None:
                self._skipped = 0
            elif self._skipped < self._skip:
                self._skipped += 1
                return False
        n_crashes = 0
        n_tries = max(self._repeat, self._min_crashes)
        if self._use_result_cache:
            with open(self.reduce_file, "rb") as test_fp:
                cache_key = hashlib.sha1(test_fp.read()).hexdigest()
            if cache_key in self._result_cache:
                result = self._result_cache[cache_key]['result']
                if result:
                    LOG.info("Interesting (cached)")
                    cached_prefix = self._result_cache[cache_key]['prefix']
                    for filename in glob.glob(r"%s_*" % cached_prefix):
                        suffix = os.path.basename(filename).split("_", 1)
                        if os.path.isfile(filename):
                            shutil.copy(filename, "%s_%s" % (temp_prefix, suffix[1]))
                        elif os.path.isdir(filename):
                            shutil.copytree(filename, "%s_%s" % (temp_prefix, suffix[1]))
                        else:
                            raise RuntimeError("Cannot copy non-file/non-directory: %s"
                                               % (filename,))
                else:
                    LOG.info("Uninteresting (cached)")
                return result

        # create the TestCase to try
        testcase = TestCase(self.landing_page, None, "grizzly.reduce", input_fname=self._input_fname)

        # add testcase contents
        for file_name in testcase_contents(self.wwwdir):
            testcase.add_from_file(os.path.join(self.wwwdir, file_name), file_name,
                                   required=bool(file_name == self.landing_page))

        # add prefs
        if self._target.prefs is not None:
            testcase.add_meta(TestFile.from_file(self._target.prefs, "prefs.js"))

        # add environment variables
        if self._env_mod is not None:
            for name, value in self._env_mod.items():
                testcase.add_environ_var(name, value)

        max_duration = 0
        run_prefix = None
        for try_num in range(n_tries):
            if (n_tries - try_num) < (self._min_crashes - n_crashes):
                break  # no longer possible to get min_crashes, so stop
            self._status.report()
            self._status.iteration += 1
            run_prefix = "%s(%d)" % (temp_prefix, try_num)
            if self._run(testcase, run_prefix):
                # track the maximum duration of the successful reduction attempts
                if testcase.duration > max_duration:
                    max_duration = testcase.duration
                n_crashes += 1
                if n_crashes >= self._min_crashes:
                    self.on_interesting_crash(run_prefix)
                    if self._use_result_cache:
                        self._result_cache[cache_key] = {
                            'result': True,
                            'prefix': run_prefix
                        }
                    self._best_testcase = testcase
                    # the amount of time it can take to replay a test case can vary
                    # when under Valgrind so do not update the timeout in that case
                    if not self._fixed_timeout and not getattr(self._target, "use_valgrind", False):
                        self.update_timeout(max_duration)
                    return True
        if self._use_result_cache:
            # No need to save the temp_prefix on uninteresting testcases
            # But let's do it anyway to stay consistent
            self._result_cache[cache_key] = {
                'result': False,
                'prefix': run_prefix
            }
        return False

    def lithium_cleanup(self):
        """Lithium cleanup entrypoint. Do any per-reduction loop cleanup needed.

        Args:
            None

        Returns:
            None
        """
        try:
            if self._server is not None:
                self._server.close()
                self._server = None
        finally:
            if self._target is not None:
                self._target.close()

    def _add_san_suppressions(self, supp_file):
        # Update the sanitizer *SAN_OPTIONS environment variable to use provided
        # suppressions file
        opt_key = '%s_OPTIONS' % os.path.basename(supp_file).split('.')[0].upper()
        opts_data = self._env_mod.get(opt_key, '')
        # the value matching *SAN_OPTIONS can be set to None
        if opts_data is None:
            opts_data = ''
        opts = sanitizer_opts(opts_data)
        opts['suppressions'] = '\'%s\'' % (supp_file,)
        self._env_mod[opt_key] = ':'.join('='.join((k, v)) for k, v in opts.items())

    def _start_log_capture(self):
        """Add a log handler for grizzly and lithium messages generated during this job.
        The handler is removed again by close()

        Args:
            None

        Returns:
            logging.Handler: The log handler to be removed later.
        """
        formatter = logging.Formatter("%(levelname).1s %(name)s [%(asctime)s] %(message)s")
        handler = logging.FileHandler(os.path.join(self._tmpdir, "reducelog.txt"))
        handler.setLevel(logging.DEBUG)
        handler.setFormatter(formatter)
        for logname in self.LOGGERS_TO_WATCH:
            logging.getLogger(logname).addHandler(handler)

        # check that DEBUG messages will actually get through
        # if the root logger level is > DEBUG, messages will not get through to our log handler
        # set root to DEBUG, and propagate the old root level to each root handler
        root_logger = logging.getLogger()
        root_level = root_logger.getEffectiveLevel()
        if root_level > logging.DEBUG:
            root_logger.setLevel(logging.DEBUG)
            for root_handler in root_logger.handlers:
                if root_handler.level < root_level:
                    root_handler.setLevel(root_level)

        return handler

    def update_timeout(self, run_time):
        # If run_time is less than poll-time, update it
        LOG.debug('Run time %r', run_time)
        new_poll_timeout = max(10, min(run_time * 1.5, self._idle_timeout))
        if new_poll_timeout < self._idle_timeout:
            LOG.info("Updating poll timeout to: %r", new_poll_timeout)
            self._idle_timeout = new_poll_timeout
        # If run_time * 2 is less than iter_timeout, update it
        # in other words, decrease the timeout if this ran in less than half the timeout
        # (floored at 10s)
        new_iter_timeout = max(10, min(run_time * 2, self._iter_timeout))
        if new_iter_timeout < self._iter_timeout:
            LOG.info("Updating max timeout to: %r", new_iter_timeout)
            self._iter_timeout = new_iter_timeout

    def _run(self, testcase, temp_prefix):
        """Run a single iteration against the target and determine if it is interesting. This is the
        low-level iteration function used by `interesting`.

        Args:
            testcase (TestCase): The testcase to serve
            temp_prefix (str): A unique prefix for any files written during this iteration.

        Returns:
            bool: True if reduced testcase is still interesting.
        """
        result = False

        # if target is closed and server is alive, we should restart it or else the first request
        #   against /first_test will 404
        if self._target.closed and self._server is not None:
            self._server.close()
            self._server = None
        self._server_map.dynamic.clear()
        self._server_map.redirect.clear()

        # launch sapphire if needed
        if self._server is None:
            # have client error pages (code 4XX) call window.close() after a few seconds
            self._server = sapphire.Sapphire(auto_close=2)

            if not self._no_harness:
                harness = os.path.join(os.path.dirname(__file__), '..', 'common', 'harness.html')
                with open(harness, 'rb') as harness_fp:
                    harness = harness_fp.read()
                self._server_map.set_dynamic_response("grz_harness", lambda: harness, mime_type="text/html")
            self._server_map.set_redirect("grz_current_test", str(self.landing_page), required=False)

        runner = Runner(self._server, self._target, self._idle_threshold, self._idle_timeout)
        if self._no_harness:
            self._server.timeout = self._iter_timeout
        else:
            # wait a few extra seconds to avoid races between the harness & sapphire timing out
            self._server.timeout = self._iter_timeout + 10

        # (re)launch Target
        if self._target.closed:
            if self._no_harness:
                location = runner.location(
                    "/grz_current_test",
                    self._server.port)
            else:
                location = runner.location(
                    "/grz_harness",
                    self._server.port,
                    close_after=self._target.rl_reset,
                    forced_close=self._target.forced_close,
                    timeout=self._iter_timeout)
            # Try to launch the browser, retry 4 times at most
            runner.launch(location, env_mod=self._env_mod, max_retries=4, retry_delay=15)
            self._target.step()

        if not self._no_harness:
            def _dyn_resp_close():  # pragma: no cover
                if self.target.monitor.is_healthy():
                    # delay to help catch window close/shutdown related crashes
                    time.sleep(0.1)
                    self.target.close()
                return b"<h1>Close Browser</h1>"
            self._server_map.set_dynamic_response("grz_close_browser", _dyn_resp_close, mime_type="text/html")
            self._server_map.set_redirect("grz_next_test", str(self.landing_page), required=True)

        # run test case
        runner.run(self._ignore, self._server_map, testcase, wait_for_callback=self._no_harness)

        # handle failure if detected
        if runner.result == Runner.FAILED:
            self._target.close()
            testcase.purge_optional(runner.served)

            # save logs
            result_logs = temp_prefix + "_logs"
            if not os.path.exists(result_logs):
                os.mkdir(result_logs)
            self._target.save_logs(result_logs)

            # create a CrashInfo
            crash = FuzzManagerReporter.create_crash_info(
                Report.from_path(result_logs),
                self._target.binary)

            short_sig = crash.createShortSignature()
            if short_sig == "No crash detected":
                # XXX: need to change this to support reducing timeouts?
                LOG.info("Uninteresting: no crash detected")
            elif self._orig_sig is None or self._orig_sig.matches(crash):
                result = True
                LOG.info("Interesting: %s", short_sig)
                if self._orig_sig is None and not self._any_crash:
                    self._orig_sig = Report.crash_signature(crash)
            else:
                LOG.info("Uninteresting: different signature: %s", short_sig)
                self.on_other_crash_found(testcase, temp_prefix)

        elif runner.result == Runner.IGNORED:
            LOG.info("Uninteresting: ignored")
            self._target.close()

        else:
            LOG.info("Uninteresting: no failure detected")

        # trigger relaunch by closing the browser if needed
        self._target.check_relaunch()

        return result

    def _stop_log_capture(self):
        """Stop handling reduce logs.

        Args:
            None

        Returns:
            None
        """
        if self._log_handler is None:
            return
        for logname in self.LOGGERS_TO_WATCH:
            logging.getLogger(logname).removeHandler(self._log_handler)
        self._log_handler.flush()
        self._log_handler.close()
        self._log_handler = None

    def config_environ(self, environ):
        with open(environ) as in_fp:
            try:
                self._env_mod = json.load(in_fp).get('env', {})
            except ValueError:
                # TODO: remove this once switched to 'test_info.json'
                # legacy support for 'env_vars.txt'
                self._env_mod = {}
                in_fp.seek(0)
                for line in in_fp:
                    line = line.rstrip()
                    if not line:
                        continue
                    key, value = line.split('=', 1)
                    if not value:
                        value = None
                    self._env_mod[key] = value
        # known sanitizer suppression files
        known_suppressions = ('lsan.supp', 'ubsan.supp')
        working_dir = os.path.dirname(environ)
        for file_name in os.listdir(working_dir):
            if file_name in known_suppressions:
                self._add_san_suppressions(os.path.join(working_dir, file_name))

    def config_signature(self, signature):
        """Configure a signature to use for reduction.  If none is given, an automatic signature is
        created based on the initial repro.

        Args:
            signature (str): A JSON signature to match for reduction.

        Returns:
            None
        """
        self._signature = CrashSignature(signature)

    @staticmethod
    def _get_landing_page(testpath):
        """Parse test_info.json for landing page

        Args:
            testpath (str): Path to a testcase folder (containing a test_info.json from Grizzly).

        Returns:
            str: Path to the landing page within testpath
        """
        info_file = os.path.join(testpath, "test_info.json")
        if os.path.isfile(info_file):
            with open(info_file) as info:
                landing_page = json.load(info).get("target", None)
            if landing_page is None:
                raise ReducerError("Could not find landing page in %s!" % (os.path.abspath(info_file),))
            landing_page = os.path.join(testpath, landing_page)
        else:
            LOG.warning("Using deprecated test_info.txt")
            with io.open(os.path.join(testpath, "test_info.txt"), encoding="utf-8") as info:
                for line in info:
                    if line.lower().startswith("landing page: "):
                        landing_page = os.path.join(testpath,
                                                    line.split(": ", 1)[1].strip())
                        break
                else:
                    raise ReducerError("Could not find landing page in %s!"
                                       % (os.path.abspath(info.name),))
        if not os.path.isfile(landing_page):
            raise ReducerError("Landing page %s does not exist in %s!"
                               % (landing_page, os.path.abspath(info.name)))
        return landing_page

    def _http_abspath(self, path):
        """Return an absolute HTTP path to `path` relative to tcroot"""
        path = os.path.relpath(path, self._tcroot)
        return '/' + '/'.join(path.split(os.sep))

    def config_testcase(self, testcase):
        """Prepare a user provided testcase for reduction.

        Args:
            testcase (str): Path to a testcase. This should be a Grizzly testcase (zip or folder) or html
                            file.

        Returns:
            None
        """
        try:
            # extract the testcase if necessary
            if os.path.exists(self._tcroot):
                raise ReducerError("Testcase already configured?")
            if os.path.isfile(testcase):
                if testcase.lower().endswith(".html"):
                    os.mkdir(self._tcroot)
                    shutil.copy(testcase, self._tcroot)
                    info = {"target": os.path.basename(testcase)}
                    with open(os.path.join(self._tcroot, "test_info.json"), "w") as info_fp:
                        json.dump(info, info_fp, indent=2, sort_keys=True)
                elif testcase.lower().endswith(".zip"):
                    os.mkdir(self._tcroot)
                    try:
                        with zipfile.ZipFile(testcase) as zip_fp:
                            zip_fp.extractall(path=self._tcroot)
                    except (zlib.error, zipfile.BadZipfile):
                        raise CorruptTestcaseError("Testcase is corrupted")
                else:
                    raise ReducerError("Testcase must be zip, html, or directory")
            elif os.path.isdir(testcase):
                shutil.copytree(testcase, self._tcroot)
            else:
                raise ReducerError("Testcase must be zip, html or directory")

            self._input_fname = os.path.basename(testcase)

            # get a list of all directories containing testcases (1-n, depending on how much history
            # grizzly saved)
            entries = set(os.listdir(self._tcroot))
            if "test_info.json" in entries:
                dirs = [self._tcroot]
            elif "test_info.txt" in entries:
                dirs = [self._tcroot]
            else:
                dirs = sorted([os.path.join(self._tcroot, entry) for entry in entries
                               if os.path.exists(os.path.join(self._tcroot, entry, "test_info.json"))
                               or os.path.exists(os.path.join(self._tcroot, entry, "test_info.txt"))],
                              key=lambda x: -int(x.rsplit('-', 1)[1]))
                if not dirs:
                    raise NoTestcaseError("No testcase recognized at %r" % (testcase,))

            # check for included prefs and environment
            if "prefs.js" in os.listdir(dirs[0]):
                # move the file out of tcroot because we prune these non-testcase files later
                os.rename(os.path.join(dirs[0], "prefs.js"), os.path.join(self._tmpdir, "prefs.js"))
                self._target.prefs = os.path.abspath(os.path.join(self._tmpdir, "prefs.js"))
                LOG.warning("Using prefs included in testcase: %r", self._target.prefs)
            if "test_info.json" in os.listdir(dirs[0]):
                self.config_environ(os.path.join(dirs[0], "test_info.json"))
            elif "env_vars.txt" in os.listdir(dirs[0]):
                # TODO: remove this block once move to 'test_info.json' is complete
                self.config_environ(os.path.join(dirs[0], "env_vars.txt"))
            if self._env_mod:
                LOG.warning("Using environment included in testcase")
                self._target.forced_close = self._env_mod.get("GRZ_FORCED_CLOSE") != "0"

            # if dirs is singular, we can use the testcase directly, otherwise we need to iterate over
            # them all in order
            pages = [self._get_landing_page(d) for d in dirs]
            if len(pages) == 1:
                self._testcase = pages[0]
                self._cache_iter_harness_created = False

            else:
                # create a harness to iterate over the whole history
                harness_path = os.path.join(os.path.dirname(__file__), '..', 'common', 'harness.html')
                with io.open(harness_path, encoding="utf-8") as harness_fp:
                    harness = harness_fp.read()
                # change dump string so that logs can be told apart
                harness = harness.replace("[grz harness]", "[cache iter]")
                # change the window name so that window.open doesn't clobber self
                harness = harness.replace("'GrizzlyFuzz'", "'CacheIterator'")
                # insert the iteration timeout. insert it directly because we can't set a hash value
                new_harness = re.sub(r"^(\s*let\s.*\btime_limit\b)",
                                     r"\1 = %d" % (self._iter_timeout * 1000),
                                     harness,
                                     flags=re.MULTILINE)
                if new_harness == harness:
                    raise ReducerError("Unable to set time_limit in harness, please update pattern "
                                       "to match harness!")
                harness = new_harness
                # make first test and next test grab from the array
                harness = harness.replace("'/grz_current_test'", "_reduce_next()")
                harness = harness.replace("'/grz_next_test'", "_reduce_next()")
                # insert the close condition. we are iterating over the array of landing pages,
                # undefined means we hit the end and the harness should close
                # newer harness uses conditional operator in open() call
                if re.search(r'open\(.*_reduce_next\(\)\s*:\s*_reduce_next\(\)', harness) is None:
                    raise ReducerError("Unable to insert finish condition, please update pattern "
                                       "to match harness!")
                # insert the landing page loop
                harness = harness.replace("<script>", "\n".join([
                    "<script>",
                    "let _reduce_tests = [",
                    "//DDBEGIN",
                    "'" + "',\n'".join(self._http_abspath(p) for p in pages) + "',",
                    "//DDEND",
                    "]",
                    "let _reduce_next = () => {",
                    "  if (!_reduce_tests.length) window.close()",
                    "  return _reduce_tests.shift()",
                    "}"
                ]))

                harness_fp, harness_path = \
                    tempfile.mkstemp(prefix="harness_", suffix=".html", dir=self._tcroot)
                os.close(harness_fp)
                with io.open(harness_path, "w", encoding="utf-8") as harness_fp:
                    harness_fp.write(harness)
                self._testcase = harness_path
                self._cache_iter_harness_created = True

            # prune unnecessary files from the testcase
            prune = {"env_vars.txt", "grizzly_fuzz_harness.html",
                     "log_metadata.json", "prefs.js", "reducelog.txt",
                     "screenlog.txt", "test_info.json", "test_info.txt"}
            for root, _, files in os.walk(self._tcroot):
                for file_ in files:
                    if file_ in prune or (file_.startswith("log_") and file_.endswith(".txt")):
                        os.unlink(os.path.join(root, file_))
        except NoTestcaseError as exc:
            LOG.warning("Could not set-up testcase: %s", exc)
            self._result_code = FuzzManagerReporter.QUAL_NO_TESTCASE
            raise
        except Exception as exc:
            LOG.warning("Could not set-up testcase: %s", exc)
            self._result_code = FuzzManagerReporter.QUAL_REDUCER_ERROR
            raise

    def __enter__(self):
        return self

    def __exit__(self, *_args):
        self.close()

    def close(self, keep_temp=False):
        """Clean up any resources used for this job.

        Args:
            None

        Returns:
            None
        """
        self._stop_log_capture()
        if self._tmpdir is not None and os.path.isdir(self._tmpdir):
            if keep_temp:
                LOG.warning("Leaving working files at %r for inspection.", self._tmpdir)
            else:
                shutil.rmtree(self._tmpdir)
                self._tmpdir = None
        if self._target is not None:
            self._target.cleanup()

    def _report_result(self, testcase, temp_prefix, quality_value, force=False):
        self._reporter.quality = quality_value
        self._reporter.force_report = force
        self._reporter.submit([testcase], log_path=temp_prefix + "_logs")

    def on_interesting_crash(self, temp_prefix):
        # called for any interesting crash
        self._interesting_prefix = temp_prefix

    def on_result(self, result_code):
        pass

    def on_other_crash_found(self, testcase, temp_prefix):
        """
        If we hit an alternate crash, store the testcase in a tmp folder.
        If the same crash is encountered again, only keep the newest one.
        """
        crash_info = FuzzManagerReporter.create_crash_info(
            Report.from_path(temp_prefix + "_logs"),
            self._target.binary)
        crash_hash = Report.crash_hash(crash_info)
        if crash_hash in self._other_crashes:
            LOG.info("Found alternate crash (newer): %s", crash_info.createShortSignature())
            # already counted when initially found
            self._status.ignored += 1
        else:
            LOG.info("Found alternate crash: %s", crash_info.createShortSignature())
            self._status.results += 1
        self._other_crashes[crash_hash] = {"tc": testcase, "prefix": temp_prefix}

    def _report_other_crashes(self):
        """
        After reduce is finished, report any alternate results (if they don't match the collector cache).
        """
        for entry in self._other_crashes.values():
            self._report_result(entry["tc"], entry["prefix"], FuzzManagerReporter.QUAL_UNREDUCED)

    def run(self, strategies=None):
        """Run reduction.
        """
        assert self._testcase is not None
        assert self._reporter is not None

        try:
            # set up lithium
            reducer = lithium.Lithium()
            self._orig_sig = self._signature
            self.landing_page = self._testcase
            reducer.conditionScript = LithiumInterestingProxy(self)

            # if we created a harness to iterate over history, files_to_reduce is initially just
            #   that harness
            # otherwise, the first stage will be skipped and we will scan for all testcases to
            #   reduce in the second stage

            run_state = RunState([self._testcase])

            # resolve list of strategies to apply
            reduce_stages = [strategies_module.MinimizeCacheIterHarness, strategies_module.ScanFilesToReduce]
            if not self._skip_analysis:
                if self._cache_iter_harness_created:
                    # if we created a cache iterator harness analyze that first
                    reduce_stages.insert(0, strategies_module.AnalyzeTestcase)
                reduce_stages.append(strategies_module.AnalyzeTestcase)
            if strategies is None:
                strategies = self.DEFAULT_STRATEGIES
            strategies_lut = strategies_module.strategies_by_name()
            for strat in strategies:
                try:
                    strat = strategies_lut[strat]
                except KeyError:
                    raise ReducerError("Unknown strategy given: %r" % (strat,))
                reduce_stages.append(strat)

            # run lithium reduce with strategies

            files_reduced = 0
            for strategy_num, strategy_type in enumerate(reduce_stages):

                result = -1
                strategy = strategy_type(self, run_state, reducer)

                for testcase_path in run_state.files_to_reduce:

                    strategy.read_testcase(testcase_path)
                    if strategy.should_skip():
                        result = 0
                        continue

                    self.reduce_file = testcase_path
                    # set up tempdir manually so it doesn't go in cwd
                    reducer.tempDir = tempfile.mkdtemp(
                        prefix="lith-%d-%s" % (strategy_num, strategy_type.name),
                        dir=self._tmpdir)

                    reducer.testCount = reducer.testTotal = 0
                    result = reducer.run()

                    try:
                        if result == 0:
                            strategy.on_success()
                            files_reduced += 1

                        else:
                            strategy.on_failure()
                            result = 0  # if we passed on failure, don't fail below

                    except StopIteration:
                        break

                if result != 0:
                    # reducer failed to repro the crash
                    if files_reduced == 0:
                        # first stage, couldn't repro at all
                        LOG.warning("Could not reduce: The testcase was not reproducible")
                        self._result_code = FuzzManagerReporter.QUAL_NOT_REPRODUCIBLE

                    else:
                        # subsequent stage, reducing broke the testcase?
                        # unclear how to recover from this.
                        # just report failure and hopefully we have another to try
                        LOG.warning("%s failed to reproduce. Previous stage broke the testcase?",
                                    strategy_type.__name__)
                        self._result_code = FuzzManagerReporter.QUAL_REDUCER_BROKE

                    return False

            # all stages succeeded
            reduced_size = run_state.total_size()
            if reduced_size == run_state.original_size:
                raise ReducerError("Reducer succeeded but nothing was reduced!")

            self._report_result(self._best_testcase,
                                self._interesting_prefix,
                                FuzzManagerReporter.QUAL_REDUCED_RESULT,
                                force=True)

            # change original quality so unbucketed crashes don't reduce again
            self._result_code = FuzzManagerReporter.QUAL_REDUCED_ORIGINAL
            return True

        except ReducerError as exc:
            LOG.warning("Could not reduce: %s", exc)
            self._result_code = FuzzManagerReporter.QUAL_REDUCER_ERROR
            return False

        except Exception:  # pylint: disable=broad-except
            LOG.exception("Exception during reduce")
            self._result_code = FuzzManagerReporter.QUAL_REDUCER_ERROR
            return False

        finally:
            self._report_other_crashes()

    def set_reporter(self, reporter):
        self._reporter = reporter

    @classmethod
    def from_args(cls, args, target, status):
        job = cls(
            args.ignore,
            target,
            args.timeout,
            args.no_harness,
            args.any_crash,
            args.skip,
            args.min_crashes,
            args.repeat,
            args.idle_threshold,
            args.idle_timeout,
            status,
            not args.no_cache,
            args.no_analysis)

        job.config_testcase(args.input)

        # arguments for environ and prefs should override the testcase
        if args.environ:
            LOG.warning("Overriding environment with %r", args.environ)
            job.config_environ(args.environ)
        if args.prefs:
            LOG.warning("Overriding prefs with %r", args.prefs)
            job.target.prefs = os.path.abspath(args.prefs)

        if args.sig is not None:
            with io.open(args.sig, encoding="utf-8") as sig_fp:
                job.config_signature(sig_fp.read())

        LOG.debug("initializing the Reporter")
        if args.fuzzmanager:
            LOG.info("Reporting issues via FuzzManager")
            job.set_reporter(FuzzManagerReporter(args.binary, tool=args.tool))
        else:
            reporter = FilesystemReporter()
            job.set_reporter(reporter)
            LOG.info("Results will be stored in %r", reporter.report_path)

        if args.fixed_timeout:
            LOG.info("Using a fixed iteration timeout")
        job._fixed_timeout = args.fixed_timeout

        return job

    @classmethod
    def main(cls, args):
        # NOTE: this mirrors grizzly.core.main pretty closely
        #       please check if updates here should go there too
        configure_logging(args.log_level)
        LOG.info("Starting Grizzly Reducer")
        if args.fuzzmanager:
            FuzzManagerReporter.sanity_check(args.binary)

        if args.ignore:
            LOG.info("Ignoring: %s", ", ".join(args.ignore))
        if args.xvfb:
            LOG.info("Running with Xvfb")
        if args.valgrind:
            LOG.info("Running with Valgrind. This will be SLOW!")

        target = None
        job = None

        status = Status.start()
        job_cancelled = False
        try:
            LOG.debug("initializing the Target")

            target = load_target(args.platform)(
                args.binary,
                args.extension,
                args.launch_timeout,
                args.log_limit,
                args.memory,
                None,  # prefs
                args.relaunch,
                valgrind=args.valgrind,
                xvfb=args.xvfb)

            job = cls.from_args(args, target, status)

            result = job.run(strategies=args.strategies)

            # report result out if callback requested
            job.on_result(job.result_code)

            # update stats
            with ReducerStats() as stats:
                if result:
                    stats.passed += 1
                elif job.result_code in (6, 10):
                    stats.failed += 1
                elif job.result_code in (7, 8, 9):
                    stats.error += 1

            if result:
                LOG.info("Reduction succeeded: %s", FuzzManagerReporter.quality_name(job.result_code))
                return Session.EXIT_SUCCESS

            LOG.warning("Reduction failed: %s", FuzzManagerReporter.quality_name(job.result_code))
            return Session.EXIT_ERROR

        except NoTestcaseError:
            with ReducerStats() as stats:
                stats.error += 1
            # TODO: test should be marked as Q7
            return Session.EXIT_ERROR

        except KeyboardInterrupt:
            job_cancelled = True
            return Session.EXIT_ABORT

        except (TargetLaunchError, TargetLaunchTimeout) as exc:
            LOG.error("Error launching target: %s", exc)
            with ReducerStats() as stats:
                stats.error += 1
            return Session.EXIT_LAUNCH_FAILURE

        finally:
            LOG.warning("Shutting down...")
            if job is not None and not job_cancelled:
                job_cancelled = job.result_code in {FuzzManagerReporter.QUAL_REDUCER_BROKE,
                                                    FuzzManagerReporter.QUAL_REDUCER_ERROR}
            if job is not None:
                job.close(keep_temp=job_cancelled)
            elif target is not None:
                # job handles calling cleanup if it was created
                target.cleanup()
            # call cleanup if we are unlikely to be using status again
            status.cleanup()



# coding=utf-8
"""
Sapphire unit tests
"""
# pylint: disable=protected-access

import hashlib
import os
import random
import threading

import pytest

from grizzly.common import TestCase

from .core import Sapphire
from .sapphire_worker import SapphireWorker
from .server_map import ServerMap
from .status_codes import SERVED_ALL, SERVED_NONE, SERVED_REQUEST, SERVED_TIMEOUT


class _TestFile(object):
    def __init__(self, url):
        self.code = None
        self.content_type = None
        self.custom_request = None
        self.len_org = 0  # original file length
        self.len_srv = 0  # served file length
        self.lock = threading.Lock()
        self.md5_org = None
        self.md5_srv = None
        self.requested = 0  # number of time file was requested
        self.url = url


def _create_test(fname, path, data=b"Test!", calc_hash=False, url_prefix=None):
    test = _TestFile(fname)
    if url_prefix is not None:
        test.url = "".join([url_prefix, fname])
    with (path / fname).open("w+b") as out_fp:
        out_fp.write(data)
        test.len_org = out_fp.tell()
        if calc_hash:
            out_fp.seek(0)
            test.md5_org = hashlib.md5(out_fp.read()).hexdigest()
    return test


def test_sapphire_00(client, tmp_path):
    """test requesting a single file"""
    with Sapphire(timeout=10) as serv:
        assert serv.timeout == 10
        test = _create_test("test_case.html", tmp_path)
        client.launch("127.0.0.1", serv.port, [test])
        assert serv.serve_path(str(tmp_path))[0] == SERVED_ALL
    assert client.wait(timeout=10)
    assert test.code == 200
    assert test.len_srv == test.len_org

def test_sapphire_01(client, tmp_path):
    """test requesting multiple files (test cleanup code)"""
    to_serve = list()
    for i in range(100):
        to_serve.append(_create_test("test_%03d.html" % i, tmp_path, data=os.urandom(5), calc_hash=True))
    with Sapphire(timeout=30) as serv:
        client.launch("127.0.0.1", serv.port, to_serve)
        status, files_served = serv.serve_path(str(tmp_path))
    assert status == SERVED_ALL
    assert len(to_serve) == len(files_served)
    assert client.wait(timeout=10)
    for t_file in to_serve:
        assert t_file.code == 200
        assert t_file.len_srv == t_file.len_org
        assert t_file.md5_srv == t_file.md5_org

def test_sapphire_02(client, tmp_path):
    """test serving optional file"""
    files_to_serve = list()
    for i in range(3):
        files_to_serve.append(_create_test("test_case_%d.html" % i, tmp_path))
    optional = [files_to_serve[0].url]
    with Sapphire(timeout=10) as serv:
        client.launch("127.0.0.1", serv.port, files_to_serve, in_order=True)
        status, served_list = serv.serve_path(str(tmp_path), optional_files=optional)
    assert status == SERVED_ALL
    assert len(files_to_serve) == len(served_list)
    assert client.wait(timeout=10)
    for t_file in files_to_serve:
        assert t_file.code == 200
        assert t_file.len_srv == t_file.len_org

def test_sapphire_03(client, tmp_path):
    """test skipping optional file"""
    files_to_serve = list()
    for i in range(3):
        files_to_serve.append(_create_test("test_case_%d.html" % i, tmp_path))
    optional = [files_to_serve[0].url]
    with Sapphire(timeout=10) as serv:
        client.launch("127.0.0.1", serv.port, files_to_serve[1:])
        status, served_list = serv.serve_path(str(tmp_path), optional_files=optional)
    assert status == SERVED_ALL
    assert len(served_list) == len(files_to_serve) - 1
    assert client.wait(timeout=10)
    assert files_to_serve[0].code is None
    assert files_to_serve[0].len_srv == 0
    for t_file in files_to_serve[1:]:
        assert t_file.code == 200
        assert t_file.len_srv == t_file.len_org

def test_sapphire_04(client, tmp_path):
    """test requesting invalid file (404)"""
    files_to_serve = list()
    files_to_serve.append(_TestFile("does_not_exist.html"))
    files_to_serve.append(_create_test("test_case.html", tmp_path))
    with Sapphire(timeout=10) as serv:
        client.launch("127.0.0.1", serv.port, files_to_serve, in_order=True)
        assert serv.serve_path(str(tmp_path))[0] == SERVED_ALL
    assert client.wait(timeout=10)
    assert "does_not_exist.html" in files_to_serve[0].url
    assert files_to_serve[0].code == 404
    assert files_to_serve[1].code == 200

def test_sapphire_05(client, tmp_path):
    """test requesting a file outside of the server root (403)"""
    files_to_serve = list()
    root_dir = tmp_path / "root"
    root_dir.mkdir()
    # add invalid file
    files_to_serve.append(_TestFile(os.path.abspath(__file__)))
    # add file in parent of root_dir
    files_to_serve.append(_create_test("no_access.html", tmp_path, data=b"no_access", url_prefix="../"))
    assert (tmp_path / "no_access.html").is_file()
    # add valid test
    files_to_serve.append(_create_test("test_case.html", root_dir))
    with Sapphire(timeout=10) as serv:
        client.launch("127.0.0.1", serv.port, files_to_serve, in_order=True)
        status, files_served = serv.serve_path(str(root_dir))
    assert status == SERVED_ALL
    assert len(files_served) == 1
    assert client.wait(timeout=10)
    assert os.path.basename(__file__) in files_to_serve[0].url
    assert files_to_serve[0].code == 403
    assert "no_access.html" in files_to_serve[1].url
    assert files_to_serve[1].code == 403
    assert files_to_serve[2].code == 200

def test_sapphire_06(client, tmp_path):
    """test serving no files... this should never happen but..."""
    with Sapphire(timeout=10) as serv:
        client.launch("127.0.0.1", serv.port, [])
        status, files_served = serv.serve_path(str(tmp_path))
    assert status == SERVED_NONE
    assert not files_served

def test_sapphire_07(tmp_path):
    """test timeout of the server"""
    with Sapphire(timeout=60) as serv:
        assert serv.timeout == 60  # verify default
        serv.timeout = None  # disable timeout
        assert serv.timeout == 0
        serv.timeout = 0  # disable timeout
        assert serv.timeout == 0
        serv.timeout = 0.1  # set minimum time
        assert serv.timeout == 1
        _create_test("test_case.html", tmp_path)
        serv._timeout = 0.01  # force shorter timeout for faster tests
        status, files_served = serv.serve_path(str(tmp_path))
    assert status == SERVED_TIMEOUT
    assert not files_served

def test_sapphire_08(client, tmp_path):
    """test only serving some files (SERVED_REQUEST)"""
    cb_status = {"count": 0}

    def is_running():
        cb_status["count"] += 1
        return cb_status["count"] < 3  # return false after 2nd call

    files_to_serve = list()
    for i in range(3):
        files_to_serve.append(_create_test("test_case_%d.html" % i, tmp_path))
    with Sapphire() as serv:
        client.launch("127.0.0.1", serv.port, files_to_serve[1:])
        status, files_served = serv.serve_path(str(tmp_path), continue_cb=is_running)
    assert status == SERVED_REQUEST
    assert len(files_served) < len(files_to_serve)

def test_sapphire_09(client, tmp_path):
    """test serving interesting sized files"""
    tests = [
        {"size": SapphireWorker.DEFAULT_TX_SIZE, "name": "even.html"},
        {"size": SapphireWorker.DEFAULT_TX_SIZE - 1, "name": "minus_one.html"},
        {"size": SapphireWorker.DEFAULT_TX_SIZE + 1, "name": "plus_one.html"},
        {"size": SapphireWorker.DEFAULT_TX_SIZE * 2, "name": "double.html"},
        {"size": 1, "name": "one.html"},
        {"size": 0, "name": "zero.html"},
    ]
    for test in tests:
        test["file"] = _TestFile(test["name"])
        t_data = "".join(random.choice("ABCD1234") for _ in range(test["size"])).encode("ascii")
        (tmp_path / test["file"].url).write_bytes(t_data)
        test["file"].md5_org = hashlib.md5(t_data).hexdigest()
    with Sapphire(timeout=10) as serv:
        client.launch("127.0.0.1", serv.port, [test["file"] for test in tests])
        status, served_list = serv.serve_path(str(tmp_path))
    assert status == SERVED_ALL
    assert len(served_list) == len(tests)
    assert client.wait(timeout=10)
    for test in tests:
        assert test["file"].code == 200
        assert test["file"].len_srv == test["size"]
        assert test["file"].md5_srv == test["file"].md5_org

def test_sapphire_10(client, tmp_path):
    """test serving a large (100MB) file"""
    t_file = _TestFile("test_case.html")
    data_hash = hashlib.md5()
    with (tmp_path / t_file.url).open("wb") as test_fp:
        # write 100MB of 'A'
        data = b"A" * (100 * 1024)  # 100KB of 'A'
        for _ in range(1024):
            test_fp.write(data)
            data_hash.update(data)
    t_file.md5_org = data_hash.hexdigest()
    with Sapphire(timeout=10) as serv:
        client.launch("127.0.0.1", serv.port, [t_file])
        assert serv.serve_path(str(tmp_path))[0] == SERVED_ALL
    assert client.wait(timeout=10)
    assert t_file.code == 200
    assert t_file.len_srv == (100 * 1024 * 1024)
    assert t_file.md5_srv == t_file.md5_org

def test_sapphire_11(client, tmp_path):
    """test serving a binary file"""
    t_file = _create_test("test_case.html", tmp_path, data=os.urandom(512), calc_hash=True)
    with Sapphire(timeout=10) as serv:
        client.launch("127.0.0.1", serv.port, [t_file])
        assert serv.serve_path(str(tmp_path))[0] == SERVED_ALL
    assert client.wait(timeout=10)
    assert t_file.code == 200
    assert t_file.len_srv == t_file.len_org
    assert t_file.md5_srv == t_file.md5_org

def test_sapphire_12():
    """test requested port is used"""
    test_port = 0x1337
    with Sapphire(port=test_port, timeout=1) as serv:
        assert test_port == serv.port

def test_sapphire_13(client, tmp_path):
    """test serving multiple content types"""
    files_to_serve = list()
    files_to_serve.append(_create_test("test_case.html", tmp_path))
    # create binary file without an extension
    files_to_serve.append(_create_test("test_case", tmp_path, data=os.urandom(5)))
    with Sapphire(timeout=10) as serv:
        client.launch("127.0.0.1", serv.port, files_to_serve)
        assert serv.serve_path(str(tmp_path))[0] == SERVED_ALL
    assert client.wait(timeout=10)
    content_types = set()
    for test in files_to_serve:
        assert test.code == 200
        assert test.len_srv == test.len_org
        file_ext = os.path.splitext(test.url)[-1]
        content_type = {".html": "text/html"}.get(file_ext, "application/octet-stream")
        content_types.add(content_type)
        assert test.content_type == content_type
    assert len(content_types) == 2

def test_sapphire_14(tmp_path):
    """test callback"""
    cb_status = {"count": 0}

    def _test_callback():
        cb_status["count"] += 1
        # return true on first call
        return cb_status["count"] < 2

    with Sapphire(timeout=10) as serv:
        _create_test("test_case.html", tmp_path)
        assert serv.serve_path(str(tmp_path), continue_cb=_test_callback)[0] == SERVED_NONE
    assert cb_status["count"] == 2

def test_sapphire_15(client, tmp_path):
    """test calling serve_path multiple times"""
    with Sapphire(timeout=10) as serv:
        for i in range(3):
            test = _create_test("test_case_%d.html" % i, tmp_path)
            client.launch("127.0.0.1", serv.port, [test])
            assert serv.serve_path(str(tmp_path))[0] == SERVED_ALL
            assert client.wait(timeout=10)
            client.close()
            assert test.code == 200
            assert test.len_srv == test.len_org
            (tmp_path / test.url).unlink()

def test_sapphire_16(client, tmp_path):
    """test non required mapped redirects"""
    smap = ServerMap()
    smap.set_redirect("test_url", "blah", required=False)
    test = _create_test("test_case.html", tmp_path)
    with Sapphire(timeout=10) as serv:
        client.launch("127.0.0.1", serv.port, [test])
        assert serv.serve_path(str(tmp_path), server_map=smap)[0] == SERVED_ALL
    assert client.wait(timeout=10)
    assert test.code == 200
    assert test.len_srv == test.len_org

def test_sapphire_17(client, tmp_path):
    """test required mapped redirects"""
    smap = ServerMap()
    with Sapphire(timeout=10) as serv:
        files_to_serve = list()
        # redir_target will be requested indirectly via the redirect
        redir_target = _create_test("redir_test_case.html", tmp_path, data=b"Redirect DATA!")
        redir_test = _TestFile("redirect_test")
        smap.set_redirect(redir_test.url, redir_target.url, required=True)
        files_to_serve.append(redir_test)
        test = _create_test("test_case.html", tmp_path)
        files_to_serve.append(test)
        client.launch("127.0.0.1", serv.port, files_to_serve)
        status, served_list = serv.serve_path(str(tmp_path), server_map=smap)
    assert status == SERVED_ALL
    assert len(served_list) == len(files_to_serve)
    assert client.wait(timeout=10)
    assert test.code == 200
    assert test.len_srv == test.len_org
    assert redir_test.code == 200
    assert redir_test.len_srv == redir_target.len_org

def test_sapphire_18(client, tmp_path):
    """test include directories and permissions"""
    inc1_path = tmp_path / "inc1"
    inc2_path = tmp_path / "inc2"
    root_path = tmp_path / "root"
    inc1_path.mkdir()
    inc2_path.mkdir()
    root_path.mkdir()
    files_to_serve = list()
    smap = ServerMap()
    with Sapphire(timeout=10) as serv:
        # add files to inc dirs
        inc1 = _create_test("included_file1.html", inc1_path, data=b"blah....1")
        files_to_serve.append(inc1)
        # add a nested dir
        nest_path = inc1_path / "nested"
        nest_path.mkdir()
        # add file in a nested dir in inc1
        nest = _create_test("nested_file.html", nest_path, data=b"blah... .nested", url_prefix="nested/")
        assert nest_path / "nested_file.html"
        files_to_serve.append(nest)
        # test 404 in nested dir in inc1
        nest_404 = _TestFile("nested/nested_file_404.html")
        files_to_serve.append(nest_404)
        # test path mounted somewhere other than /
        inc2 = _create_test("included_file2.html", inc2_path, data=b"blah....2", url_prefix="inc_test/")
        files_to_serve.append(inc2)
        # test 404 in include dir
        inc404 = _TestFile("inc_test/included_file_404.html")
        assert not (nest_path / "included_file_404.html").is_file()
        files_to_serve.append(inc404)
        # test 403
        inc403 = _create_test("no_access.html", tmp_path, data=b"no_access", url_prefix="inc_test/../")
        assert (tmp_path / "no_access.html").is_file()
        files_to_serve.append(inc403)
        # test file (used to keep sever job alive)
        test = _create_test("test_case.html", root_path)
        files_to_serve.append(test)
        # add include paths
        smap.set_include("/", str(inc1_path))  # mount at '/'
        smap.set_include("inc_test", str(inc2_path))  # mount at '/inc_test'
        client.launch("127.0.0.1", serv.port, files_to_serve, in_order=True)
        status, files_served = serv.serve_path(str(root_path), server_map=smap)
    assert status == SERVED_ALL
    assert "test_case.html" in files_served
    assert str(inc1_path / "included_file1.html") in files_served
    assert str(inc2_path / "included_file2.html") in files_served
    assert str(nest_path / "nested_file.html") in files_served
    assert client.wait(timeout=10)
    assert inc1.code == 200
    assert inc2.code == 200
    assert nest.code == 200
    assert test.code == 200
    assert nest_404.code == 404
    assert inc404.code == 404
    assert inc403.code == 403

def test_sapphire_19(client, tmp_path):
    """test dynamic response"""
    _test_string = b"dynamic response -- TEST DATA!"

    def _dyn_test_cb():
        return _test_string

    smap = ServerMap()
    test_dr = _TestFile("dynm_test")
    test_dr.len_org = len(_test_string)
    test_dr.md5_org = hashlib.md5(_test_string).hexdigest()
    smap.set_dynamic_response("dynm_test", _dyn_test_cb, mime_type="text/plain")
    test = _create_test("test_case.html", tmp_path)
    with Sapphire(timeout=10) as serv:
        client.launch("127.0.0.1", serv.port, [test_dr, test], in_order=True)
        assert serv.serve_path(str(tmp_path), server_map=smap)[0] == SERVED_ALL
    assert client.wait(timeout=10)
    assert test.code == 200
    assert test.len_srv == test.len_org
    assert test_dr.code == 200
    assert test_dr.len_srv == test_dr.len_org
    assert test_dr.md5_srv == test_dr.md5_org

def test_sapphire_20(client_factory, tmp_path):
    """test pending_files == 0 in worker thread"""
    client_defer = client_factory(rx_size=2)
    # server should shutdown while this file is being served
    test_defer = _create_test("defer_test.html", tmp_path)
    optional = [test_defer.url]
    test = _create_test("test_case.html", tmp_path, data=b"112233")
    with Sapphire(timeout=10) as serv:
        # this test needs to wait just long enough to have the required file served
        # but not too long or the connection will be closed by the server
        client_defer.launch("127.0.0.1", serv.port, [test_defer], delay=0.1, indicate_failure=True)
        client = client_factory(rx_size=2)
        client.launch("127.0.0.1", serv.port, [test], throttle=0.1)
        assert serv.serve_path(str(tmp_path), optional_files=optional)[0] == SERVED_ALL
    assert client_defer.wait(timeout=10)
    assert client.wait(timeout=10)
    assert test.code == 200
    assert test_defer.code == 0

def test_sapphire_21(client, tmp_path):
    """test handling an invalid request"""
    bad_test = _TestFile("bad.html")
    bad_test.custom_request = b"a bad request...0+%\xef\xb7\xba\r\n"
    optional = [bad_test.url]
    test = _create_test("test_case.html", tmp_path)
    with Sapphire(timeout=10) as serv:
        client.launch("127.0.0.1", serv.port, [bad_test, test], in_order=True)
        assert serv.serve_path(str(tmp_path), optional_files=optional)[0] == SERVED_ALL
    assert client.wait(timeout=10)
    assert test.code == 200
    assert bad_test.code == 400

def test_sapphire_22(client, tmp_path):
    """test handling an empty request"""
    bad_test = _TestFile("bad.html")
    bad_test.custom_request = b""
    optional = [bad_test.url]
    test = _create_test("test_case.html", tmp_path)
    with Sapphire(timeout=10) as serv:
        client.launch("127.0.0.1", serv.port, [bad_test, test], indicate_failure=True, in_order=True)
        assert serv.serve_path(str(tmp_path), optional_files=optional)[0] == SERVED_ALL
    assert client.wait(timeout=10)
    assert test.code == 200
    assert bad_test.code == 0

def test_sapphire_23(client_factory, tmp_path):
    """test requesting multiple files via multiple connections"""
    to_serve = list()
    for i in range(2):
        to_serve.append(_create_test("test_%03d.html" % i, tmp_path, data=b"AAAA"))
    max_workers = 20
    with Sapphire(max_workers=max_workers, timeout=60) as serv:
        clients = list()
        try:
            for _ in range(max_workers):  # number of clients to spawn
                clients.append(client_factory(rx_size=1))
            for client in clients:
                client.launch("127.0.0.1", serv.port, to_serve, in_order=True, throttle=0.05)
            status, files_served = serv.serve_path(str(tmp_path))
            # call serv.close() instead of waiting for the clients to timeout
            serv.close()
        finally:
            for client in clients:
                assert client.wait(timeout=10)
                client.close()
    assert status == SERVED_ALL
    assert len(to_serve) == len(files_served)
    for t_file in to_serve:
        assert t_file.code == 200
        assert t_file.len_srv == t_file.len_org

def test_sapphire_24(client_factory, tmp_path):
    """test all request types via multiple connections"""
    def _dyn_test_cb():
        return b"A" if random.getrandbits(1) else b"AA"

    smap = ServerMap()
    with Sapphire(max_workers=10, timeout=60) as serv:
        to_serve = list()
        for i in range(50):
            # add required files
            to_serve.append(_create_test("test_%03d.html" % i, tmp_path, data=b"A" * ((i % 2) + 1)))
            # add a missing files
            to_serve.append(_TestFile("missing_%03d.html" % i))
            # add optional files
            opt_path = tmp_path / ("opt_%03d.html" % i)
            opt_path.write_bytes(b"A" * ((i % 2) + 1))
            to_serve.append(_TestFile(opt_path.name))
            # add redirects
            redir_target = _create_test("redir_%03d.html" % i, tmp_path, data=b"AA")
            to_serve.append(_TestFile("redir_%03d" % i))
            smap.set_redirect(to_serve[-1].url, redir_target.url, required=random.getrandbits(1) > 0)
            # add dynamic responses
            to_serve.append(_TestFile("dynm_%03d" % i))
            smap.set_dynamic_response(to_serve[-1].url, _dyn_test_cb, mime_type="text/plain")
        clients = list()
        for _ in range(100):  # number of clients to spawn
            clients.append(client_factory(rx_size=1))
            throttle = 0.05 if random.getrandbits(1) else 0
            clients[-1].launch("127.0.0.1", serv.port, to_serve, throttle=throttle)
        assert serv.serve_path(str(tmp_path), server_map=smap)[0] == SERVED_ALL

def test_sapphire_25(client, tmp_path):
    """test dynamic response with bad callbacks"""
    test_dr = _TestFile("dynm_test")
    smap = ServerMap()
    smap.set_dynamic_response("dynm_test", lambda: None, mime_type="text/plain")
    test = _create_test("test_case.html", tmp_path)
    with Sapphire(timeout=10) as serv:
        client.launch("127.0.0.1", serv.port, [test_dr, test], in_order=True)
        with pytest.raises(TypeError):
            serv.serve_path(str(tmp_path), server_map=smap)

def test_sapphire_26(client, tmp_path):
    """test serving to a slow client"""
    t_data = "".join(random.choice("ABCD1234") for _ in range(0x19000))  # 100KB
    t_file = _create_test("test_case.html", tmp_path, data=t_data.encode("ascii"), calc_hash=True)
    # rx_size 10KB and throttle to 0.25 sec, which will be ~50KB/s
    # also taking 2.5 seconds to complete will hopefully find problems
    # with any assumptions that were made
    client.rx_size = 0x2800
    with Sapphire(timeout=60) as serv:
        client.launch("127.0.0.1", serv.port, [t_file], throttle=0.25)
        assert serv.serve_path(str(tmp_path))[0] == SERVED_ALL
    assert client.wait(timeout=10)
    assert t_file.code == 200
    assert t_file.len_srv == t_file.len_org
    assert t_file.md5_srv == t_file.md5_org

def test_sapphire_27(client, tmp_path):
    """test timeout while requesting multiple files"""
    files_to_serve = list()
    t_data = "".join(random.choice("ABCD1234") for _ in range(1024)).encode("ascii")
    for i in range(50):
        files_to_serve.append(_create_test("test_case_%03d.html" % i, tmp_path, data=t_data))
    client.rx_size = 512
    with Sapphire(timeout=1) as serv:  # minimum timeout is 1 second
        client.launch("127.0.0.1", serv.port, files_to_serve, indicate_failure=True, throttle=0.1)
        status, files_served = serv.serve_path(str(tmp_path))
    assert status == SERVED_TIMEOUT
    assert len(files_served) < len(files_to_serve)

def test_sapphire_28(client, tmp_path):
    """test Sapphire.serve_testcase()"""
    with TestCase("test.html", "none.test", "foo") as test:
        test.add_from_data(b"test", "test.html")
        t_file = _create_test(test.landing_page, tmp_path)
        with Sapphire(timeout=10) as serv:
            client.launch("127.0.0.1", serv.port, [t_file])
            assert test.duration is None
            status, files_served = serv.serve_testcase(test)
    assert status == SERVED_ALL
    assert files_served
    assert test.duration >= 0

def test_sapphire_29(client_factory, tmp_path):
    """test Sapphire.serve_path() with forever=True"""
    clients = list()
    with Sapphire(timeout=10) as serv:
        assert serv.timeout == 10
        test = _create_test("test_case.html", tmp_path)
        for _ in range(3):
            clients.append(client_factory())
        for client in clients:
            client.launch("127.0.0.1", serv.port, [test], skip_served=False)
        def _test_callback():
            with test.lock:
                return test.requested < 3
        assert serv.serve_path(str(tmp_path), continue_cb=_test_callback, forever=True)[0] == SERVED_ALL
    for client in clients:
        assert client.wait(timeout=10)
        client.close()
    assert test.requested == 3
    assert test.code == 200
    assert test.len_srv == test.len_org

def test_sapphire_30(client, tmp_path):
    """test interesting file names"""
    to_serve = [
        # space in file name
        _create_test("test case.html", tmp_path),
        # non-alphanumeric chars (valid characters to use on filesystem)
        _create_test("!@#$%^&(_+-=[]),;'~`{}", tmp_path)]
    with Sapphire(timeout=10) as serv:
        client.launch("127.0.0.1", serv.port, to_serve)
        assert serv.serve_path(str(tmp_path))[0] == SERVED_ALL
    assert client.wait(timeout=10)
    for t_file in to_serve:
        assert t_file.code == 200

def test_sapphire_31(client, tmp_path):
    """test interesting path string"""
    all_bytes = "".join(chr(i) for i in range(256))
    to_serve = [
        # should not trigger crash
        _TestFile(all_bytes),
        # used to keep server running
        _create_test("a.html", tmp_path)]
    with Sapphire(timeout=10) as serv:
        client.launch("127.0.0.1", serv.port, to_serve, in_order=True)
        assert serv.serve_path(str(tmp_path), optional_files=[all_bytes])[0] == SERVED_ALL
    assert client.wait(timeout=10)
    for t_file in to_serve:
        assert t_file.code is not None

def test_sapphire_32(mocker):
    """test Sapphire._create_listening_socket()"""
    fake_sleep = mocker.patch("sapphire.core.time.sleep", autospec=True)
    fake_sock = mocker.patch("sapphire.core.socket.socket", autospec=True)
    assert Sapphire._create_listening_socket(False, None)
    assert fake_sock.return_value.close.call_count == 0
    assert fake_sock.return_value.setsockopt.call_count == 1
    assert fake_sock.return_value.settimeout.call_count == 1
    assert fake_sock.return_value.bind.call_count == 1
    assert fake_sock.return_value.listen.call_count == 1
    assert fake_sleep.call_count == 0
    fake_sock.reset_mock()
    # failure to bind
    fake_sock.return_value.bind.side_effect = OSError
    with pytest.raises(OSError):
        Sapphire._create_listening_socket(False, None)
    assert fake_sock.return_value.close.call_count == 1
    assert fake_sleep.call_count == 0
    fake_sock.reset_mock()
    # failure and pass on retry
    exc = OSError()
    exc.errno = 10013
    fake_sock.return_value.bind.side_effect = (exc, None)
    assert Sapphire._create_listening_socket(False, None)
    assert fake_sock.return_value.close.call_count == 1
    assert fake_sock.return_value.listen.call_count == 1
    assert fake_sleep.call_count == 1

def test_main_01(mocker, tmp_path):
    """test Sapphire.main()"""
    args = mocker.Mock(
        path=str(tmp_path),
        port=4536,
        remote=False,
        timeout=None)
    Sapphire.main(args)



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from abc import ABCMeta, abstractmethod, abstractproperty
from os import remove


__all__ = ("TargetMonitor",)
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith", "Jesse Schwartzentruber"]


class TargetMonitor(metaclass=ABCMeta):
    @abstractmethod
    def clone_log(self, log_id, offset=0):
        pass

    @abstractmethod
    def is_healthy(self):
        pass

    @abstractmethod
    def is_running(self):
        pass

    @abstractproperty
    def launches(self):
        pass

    def log_data(self, log_id, offset=0):
        log_file = self.clone_log(log_id, offset=offset)
        if log_file is None:
            return None
        try:
            with open(log_file, "rb") as log_fp:
                return log_fp.read()
        finally:
            remove(log_file)

    @abstractmethod
    def log_length(self, log_id):
        pass
